import threading, copy, logging
from datetime import datetime
from ..singleton import Singleton
from ..models import FloatType, StringType, IntegerType, BooleanType
from ..modules.users.users import User
from ..modules.users.users import User
from ..utils.decorators import set_event, logging_error_handler
from ..filter import filter
# from ..iad import iad_outlier, iad_frozen_data, iad_out_of_range
from .tag import Tag
from flask_socketio import SocketIO

class CVT:
    """
    Current Value Table (CVT) class for a tag-based repository.

    This class is designed to hold in-memory tag-based values and manage observers for the required tags. It is intended to be used exclusively by PyAutomation and should not be used for other purposes.

    **Usage Example**:

    .. code-block:: python

        >>> from automation.tags import CVT
        >>> _cvt = CVT()

    """

    def __init__(self):
        
        self._tags = dict()
        self.data_types = ["float", "int", "bool", "str"]
        self.sio:SocketIO|None = None

    def set_socketio(self, sio:SocketIO):
        r"""
        Documentation here
        """
        self.sio:SocketIO = sio

    @set_event(message=f"Created", classification="Tag", priority=1, criticity=1)
    def set_tag(
        self, 
        name:str, 
        unit:str, 
        data_type:str, 
        description:str, 
        variable:str,
        display_name:str="",
        display_unit:str="",
        opcua_address:str="",
        node_namespace:str="",
        scan_time:int=None,
        dead_band:float=None,
        process_filter:bool=False,
        gaussian_filter:bool=False,
        gaussian_filter_threshold:float=1.0,
        gaussian_filter_r_value:float=0.0,
        outlier_detection:bool=False,
        out_of_range_detection:bool=False,
        frozen_data_detection:bool=False,
        manufacturer:str="",
        segment:str="",
        id:str=None,
        user:User=None
        )->tuple[Tag, str]:
        """Initialize a new Tag object in the _tags dictionary.
        
        # Parameters
        name (str):
            Tag name.
        data_type (str): 
            Tag value type ("int", "float", "bool", "str")
        """
        if isinstance(data_type, str):
        
            if data_type in self.data_types:
                if data_type == "float":
                    value = 0.0
                elif data_type == "int":
                    value = 0
                elif data_type == "str":
                    value = ""
                else:
                    value = False
            
        elif isinstance(data_type, (FloatType, IntegerType, StringType, BooleanType)):

            value = data_type()
            data_type.set(name, value)
            data_type = data_type.__name__
            self.set_data_type(data_type)
        
        has_duplicates, message = self.has_duplicates(name=name, display_name=display_name, opcua_address=opcua_address, node_namespace=node_namespace)
        if has_duplicates:

            return None, message
        
        if not display_unit:

            display_unit = unit

        
        tag = Tag(
            name=name,
            unit=unit,
            data_type=data_type,
            description=description,
            variable=variable,
            display_name=display_name,
            display_unit=display_unit,
            opcua_address=opcua_address,
            node_namespace=node_namespace,
            scan_time=scan_time,
            dead_band=dead_band,
            process_filter=process_filter,
            gaussian_filter=gaussian_filter,
            gaussian_filter_threshold=gaussian_filter_threshold,
            gaussian_filter_r_value=gaussian_filter_r_value,
            outlier_detection=outlier_detection,
            out_of_range_detection=out_of_range_detection,
            frozen_data_detection=frozen_data_detection,
            manufacturer=manufacturer,
            segment=segment,
            id=id
        )
        self._tags[tag.id] = tag

        return tag, message

    @set_event(message=f"Updated", classification="Tag", priority=1, criticity=3)
    def update_tag(
        self, 
        id:str,  
        user:User=None, 
        **kwargs
        )->tuple[Tag|None, str]:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        check = dict()
        if "name" in kwargs:
            check["name"] = kwargs["name"]
        if "display_name" in kwargs:
            check["display_name"] = kwargs["display_name"]
        if "node_namespace" in kwargs:
            check["node_namespace"] = kwargs["node_namespace"]
        if "opcua_address" in kwargs:
            check["opcua_address"] = kwargs["opcua_address"]
        has_duplicates, message = self.has_duplicates(**check)
        if has_duplicates:

            return None, message
        
        tag = self._tags[id]
        if "name" in kwargs:
            tag.set_name(name=kwargs["name"])
        if "unit" in kwargs:
            tag.set_unit(unit=kwargs["unit"])
        if "data_type" in kwargs:
            tag.set_data_type(data_type=kwargs["data_type"])
        if "description" in kwargs:
            tag.set_description(description=kwargs["description"])
        if "variable" in kwargs:
            tag.set_variable(variable=kwargs["variable"])
        if "display_name" in kwargs:
            tag.set_display_name(name=kwargs["display_name"])
        if "display_unit" in kwargs:
            tag.set_display_unit(unit=kwargs["display_unit"])
        if "opcua_address" in kwargs:
            tag.set_opcua_address(opcua_address=kwargs["opcua_address"])
        if "node_namespace" in kwargs:
            tag.set_node_namespace(node_namespace=kwargs["node_namespace"])
        if "scan_time" in kwargs:
            if isinstance(kwargs["scan_time"], int):
                tag.set_scan_time(scan_time=kwargs["scan_time"])
        if "dead_band" in kwargs:
            tag.set_dead_band(dead_band=kwargs["dead_band"])
        if "segment" in kwargs:
            tag.segment = kwargs["segment"]
        if "manufacturer" in kwargs:
            tag.manufacturer = kwargs["manufacturer"]
        if "gaussian_filter" in kwargs:
            
            if kwargs['gaussian_filter'].lower() in ('1', 'true'):

                tag.gaussian_filter = True

            else:

                tag.gaussian_filter = False

        if "gaussian_filter_r_value" in kwargs:

            tag.gaussian_filter_r_value = kwargs['gaussian_filter_r_value']

        if "gaussian_filter_threshold" in kwargs:

            tag.gaussian_filter_threshold = kwargs['gaussian_filter_threshold']
        
        self._tags[id] = tag

        return tag, f"Tag: {tag.name}"

    @set_event(message=f"Updated", classification="Tag", priority=1, criticity=5)
    def delete_tag(self, id:str, user:User):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        tag = self._tags.pop(id)
        return tag, f"Tag: {tag.name}"

    def get_tag(self, id:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _id, tag in self._tags.items():

            if _id==id:
                
                return tag

        return None
    
    def get_unit_by_tag(self, tag:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _id, _tag in self._tags.items():

            if _tag.name==tag:
                
                return _tag.unit

        return None
    
    def get_display_unit_by_tag(self, tag:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _id, _tag in self._tags.items():
            
            if _tag.name==tag:
                
                return _tag.display_unit

        return None

    def get_tags(self)->list:
        r"""
        Returns a list of the defined tags names.
        """
        if self._tags:

            return [tag.serialize() for _, tag in self._tags.items()]
        
        return list()
    
    def get_tags_by_names(self, names:list)->list:
        r"""
        Returns a list of the defined tags names.
        """
        if self._tags:

            return [tag.serialize() for _, tag in self._tags.items() if tag.name in names]
        
        return list()
    
    def get_field_tags_names(self)->list:
        r"""
        Returns a list of the defined tags names.
        """
        if self._tags:

            return [tag.name for _, tag in self._tags.items() if tag.opcua_address and tag.node_namespace]
        
        return list()
    
    def get_cuasi_field_tags_names(self)->list:
        r"""
        Returns a list of the defined tags names.
        """
        if self._tags:

            return [tag.name for _, tag in self._tags.items() if tag.opcua_address]
        
        return list()
    
    def get_tag_by_name(self, name:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _, tag in self._tags.items():
            
            if tag.get_name()==name:
                
                return tag

        return None
    
    def get_tag_by_display_name(self, display_name:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _, tag in self._tags.items():

            if tag.get_display_name()==display_name:
                
                return tag

        return None

    def get_tag_by_node_namespace(self, node_namespace:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        for _, tag in self._tags.items():

            if tag.get_node_namespace()==node_namespace:
                
                return tag

        return None
    
    def get_value(self, id:str)->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        tag = self._tags[id]        
        _new_object = copy.copy(tag.get_value())
        return _new_object
    
    def get_timestamp(self, id:str)->datetime:
        r"""
        Documentation here
        """
        tag = self._tags[id] 

        return tag.get_timestamp()
    
    def get_value_by_name(self, name:str)->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """

        tag = self.get_tag_by_name(name=name)  

        return {
                "value": tag.get_value(),
                "unit": tag.get_unit(),
                "timestamp": tag.get_timestamp()
            }

    def get_values_by_name(self, names:list[str])->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        data = dict()

        for name in names:

            tag = self.get_tag_by_name(name=name)  
            data[name] = {
                "value": tag.get_value(),
                "unit": tag.get_unit(),
                "timestamp": tag.get_timestamp()
            }
            
        return data
    
    @logging_error_handler
    @filter
    # @iad_frozen_data
    # @iad_out_of_range
    # @iad_outlier
    def set_value(self, id:str, value, timestamp:datetime):
        """Sets a new value for a defined tag.
        
        # Parameters
        name (str):
            Tag name.
        value (float, int, bool): 
            Tag value ("int", "float", "bool")
        """
        from .. import TIMEZONE
        tag = self._tags[id]
        
        # Deadband Logic Wrapper for CVT
        

        if tag.dead_band and isinstance(value, (int, float)):
            try:
                current_value = tag.value.value
                if abs(value - current_value) < tag.dead_band:
                    return value
            except Exception as e:
                logging.error(f"Error in deadband logic: {e}")

        tag.set_value(value=value, timestamp=timestamp)
        if self.sio:
            timestamp = timestamp.astimezone(TIMEZONE)
            self._tags[id].timestamp = timestamp
            self.sio.emit("on.tag", data=self._tags[id].serialize())

        return value

    def set_data_type(self, data_type):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        self.data_types.append(data_type)
        self.data_types = list(set(self.data_types))
    
    def is_tag_defined(self, name:str)->bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """

        return name in self._tags
    
    def attach_observer(self, name, observer):
        r"""Attaches a new observer to a tag object defined by name.
        
        # Parameters
        name (str):
            Tag name.
        observer (TagObserver): 
            Tag observer object, will update once a tag object is changed.
        """
        tag = self.get_tag_by_name(name)
        if tag:
            
            self._tags[tag.id].attach(observer)
        
        else:
            logger = logging.getLogger("pyautomation")
            logger.warning(f"{name} tag Not exists in CVT.attach_observer method")

    def detach_observer(self, name, observer):
        r"""
        Detaches an observer from a tag object defined by name.
        
        # Parameters
        name (str):
            Tag name.
        observer (TagObserver): 
            Tag observer object.
        """
        tag = self.get_tag_by_name(name)
        self._tags[tag.id].detach(observer)

    def has_duplicates(self, tag:Tag=None, name:str=None, display_name:str=None, node_namespace:str=None, opcua_address:str=None):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """

        for _, _tag in self._tags.items():

            if name:

                if _tag.get_name()==name:

                    return True, f"Duplicated Tag Name: {name}"
                
            if display_name:
            
                if _tag.get_display_name()==display_name:

                    return True, f"Duplicated Display Name: {display_name}"
                
            if node_namespace:
            
                if _tag.get_node_namespace()==node_namespace:
                    
                    if tag:
                    
                        if _tag.get_opcua_address()==tag.get_opcua_address():

                            return True, f"Duplicated Node Namespace: {node_namespace}"
                        
                    return True, f"Duplicated Node Namespace: {node_namespace}"
            
        return False, f"Valid Tag Name: {name} - Display Name: {display_name}"

    def serialize(self, id:str)->dict:
        r"""Returns a tag type defined by name.
        
        # Parameters
        name (str):
            Tag name.
        """
        return self._tags[id].serialize()
    
    def serialize_by_tag_name(self, name:str)->dict|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        tag = self.get_tag_by_name(name)

        if tag:

            return tag.serialize()


class CVTEngine(Singleton):
    """
    Current Value Table (CVT) Engine class for a tag-based, thread-safe repository.

    This class is designed to hold in-memory tag-based values and manage observers for the required tags. It is implemented as a singleton, ensuring that each sub-thread within the PyAutomation application can access and modify tags in a thread-safe manner.

    **Usage Example**:

    .. code-block:: python

        >>> from automation.tags import CVTEngine
        >>> tag_engine = CVTEngine()
    """


    def __init__(self):

        super(CVTEngine, self).__init__()
        self._cvt = CVT()
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()
        self._config = None
        self._response = None
        self._response_lock.acquire()
        self.DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"

    def set_tag(
        self, 
        name:str, 
        unit:str, 
        data_type:str, 
        variable:str,
        description:str, 
        display_unit:str="",
        display_name:str="",
        opcua_address:str="",
        node_namespace:str="",
        scan_time:int=0,
        dead_band:float=0.0,
        process_filter:bool=False,
        gaussian_filter:bool=False,
        gaussian_filter_threshold:float=1.0,
        gaussian_filter_r_value:float=0.0,
        outlier_detection:bool=False,
        out_of_range_detection:bool=False,
        frozen_data_detection:bool=False,
        manufacturer:str="",
        segment:str="",
        id:str="",
        user:User|None=None
        )->tuple[Tag, str]:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "set_tag"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["unit"] = unit
        _query["parameters"]["data_type"] = data_type
        _query["parameters"]["variable"] = variable
        _query["parameters"]["description"] = description
        _query["parameters"]["display_unit"] = display_unit
        _query["parameters"]["display_name"] = display_name
        _query["parameters"]["opcua_address"] = opcua_address
        _query["parameters"]["node_namespace"] = node_namespace
        _query["parameters"]["scan_time"] = scan_time
        _query["parameters"]["dead_band"] = dead_band
        _query["parameters"]["process_filter"] = process_filter
        _query["parameters"]["gaussian_filter"] = gaussian_filter
        _query["parameters"]["gaussian_filter_threshold"] = gaussian_filter_threshold
        _query["parameters"]["gaussian_filter_r_value"] = gaussian_filter_r_value
        _query["parameters"]["outlier_detection"] = outlier_detection
        _query["parameters"]["out_of_range_detection"] = out_of_range_detection
        _query["parameters"]["frozen_data_detection"] = frozen_data_detection
        _query["parameters"]["manufacturer"] = manufacturer
        _query["parameters"]["segment"] = segment
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        return self.__query(_query)
    
    def update_tag(
            self, 
            id:str,  
            user:User=None, 
            **kwargs
        ):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "update_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        for key, value in kwargs.items():

            _query["parameters"][key] = value
        return self.__query(_query)
    
    def delete_tag(self, id:str, user:User|None=None):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "delete_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["user"] = user
        return self.__query(_query)
    
    def get_tag(
        self,
        id:str=None
        )->Tag:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tag"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)

    def get_tags(self):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tags"
        return self.__query(_query)

    def get_tags_by_names(self, names:list[str]):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tags_by_names"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = names
        return self.__query(_query)
    
    def get_tag_by_name(self, name:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tag_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)
    
    def get_tag_by_display_name(self, display_name:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tag_by_display_name"
        _query["parameters"] = dict()
        _query["parameters"]["display_name"] = display_name
        return self.__query(_query)

    def get_tag_by_node_namespace(self, node_namespace:str)->Tag|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_tag_by_node_namespace"
        _query["parameters"] = dict()
        _query["parameters"]["node_namespace"] = node_namespace
        return self.__query(_query)
    
    def get_value(self, id:str)->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_value"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_value_by_name(self, tag_name:str)->dict:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_value_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = tag_name
        return self.__query(_query)
    
    def get_values_by_name(self, tag_names:list[str])->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_values_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = tag_names
        return self.__query(_query)
    
    def get_scan_time(self, id:str)->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_scan_time"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_dead_band(self, id:str)->str|float|int|bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_dead_band"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)
    
    def get_display_unit_by_tag(self, tag:str)->str:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "get_display_unit_by_tag"
        _query["parameters"] = dict()
        _query["parameters"]["tag"] = tag
        return self.__query(_query)
    
    @logging_error_handler
    def set_value(self, id:str, value, timestamp:datetime):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        if not timestamp:
            timestamp = datetime.now()
        _query = dict()
        _query["action"] = "set_value"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["value"] = value
        _query["parameters"]["timestamp"] = timestamp
        return self.__query(_query)
    
    def set_data_type(self, data_type):
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "set_data_type"
        _query["parameters"] = dict()
        _query["parameters"]["data_type"] = data_type
        return self.__query(_query)
    
    def is_tag_defined(self, name:str)->bool:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "is_tag_defined"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)

    def attach(self, name:str, observer):
        """
        Attaches an observer object to a Tag, observer gets notified when the Tag value changes.
        
        **Parameters:**

        * **name** (str): Tag name.
        * **observer** (str): TagObserver instance.
        """
        _query = dict()
        _query["action"] = "attach_observer"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["observer"] = observer
        return self.__query(_query)

    def detach(self, name:str, observer):
        """
        Detaches an observer object from a Tag, observer no longer gets notified when the Tag value changes.
        
        **Parameters:**

        * **name** (str): Tag name.
        * **observer** (str): TagObserver instance.
        """
        
        _query = dict()
        _query["action"] = "detach_observer"

        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["observer"] = observer

        self.request(_query)
        result = self.response()

        if result["result"]:
            return result["response"]

    def serialize(self, id:str)->dict:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "serialize"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.__query(_query)

    def serialize_by_tag_name(self, name:str)->dict|None:
        r"""Documentation here

        # Parameters

        - 

        # Returns

        - 
        """
        _query = dict()
        _query["action"] = "serialize_by_tag_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        return self.__query(_query)

    def __query(self, query:dict)->dict:

        self.request(query)
        result = self.response()
        if result["result"]:
            return result["response"]

    def request(self, query:dict):
        r"""
        It does the request to the tags repository according query's structure, in a thread-safe mechanism

        **Parameters**

        * **query** (dict): Query to tags repository

        ## Query Structure

        ```python
        query = {
            "action": (str)
            "parameters": (dict)
        }
        ```
        ## Valid actions in query

        * set_tag
        * get_tags
        * get_value
        * get_data_type
        * get_unit
        * get_description
        * get_display_name
        * get_min_value
        * get_max_value
        * get_attributes
        * set_value
        * attach
        * detach

        ## Parameters strcuture in query

        ```python
        parameters = {
            "name": (str) tag name to do request
            "unit": (str)[Optional] Unit to get value
            "value": (float)[Optional] If you use *set_value* function, you must pass this parameter
            "observer": (TagObserver)[Optional] If you use *attach* and *detach* function, you must pass this parameter
        }
        ```

        """
        self._request_lock.acquire()
        action = query["action"]
        error_msg = f"Error in CVTEngine with action: {action}"

        try:

            if hasattr(self._cvt, action):

                method = getattr(self._cvt, action)
                
                if 'parameters' in query:
                    
                    resp = method(**query["parameters"])

                else:

                    resp = method()

            self.__true_response(resp)

        except Exception as e:
            
            self.__log_error(e, error_msg)

        self._response_lock.release()

    def __log_error(self, e:Exception, msg:str):
        r"""
        Documentation here
        """
        logging.error(f"{e} Message: {msg}")
        self._response = {
            "result": False,
            "response": None
        }

    def __true_response(self, resp):
        r"""
        Documentation here
        """
        self._response = {
            "result": True,
            "response": resp
        }

    def response(self)->dict:
        r"""
        Handles the python GIL to emit the request's response in a thread-safe mechanism.
        """
        self._response_lock.acquire()

        result = self._response

        self._request_lock.release()

        return result

    def __getstate__(self):

        self._response_lock.release()
        state = self.__dict__.copy()
        del state['_request_lock']
        del state['_response_lock']
        return state

    def __setstate__(self, state):
        
        self.__dict__.update(state)
        self._request_lock = threading.Lock()
        self._response_lock = threading.Lock()
        self._response_lock.acquire()

    


# -*- coding: utf-8 -*-
"""pyhades/logger/events.py
"""
from datetime import datetime
from ..dbmodels.events import Events
from ..modules.users.users import User
from .core import BaseEngine, BaseLogger


class EventsLogger(BaseLogger):

    def __init__(self):

        super(EventsLogger, self).__init__()

    def create(
        self, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        ):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return None
            
        return Events.create(
            message=message, 
            user=user, 
            description=description, 
            classification=classification,
            priority=priority,
            criticity=criticity,
            timestamp=timestamp
        )

    def get_lasts(self, lasts:int=1):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return list()
        
        if not self.check_connectivity():
            
            return list()
        
        return Events.read_lasts(lasts=lasts)
    
    def filter_by(
        self,
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        message:str="",
        description:str="",
        classification:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str="UTC",
        page:int=1,
        limit:int=20
        ):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return Events.filter_by(
            usernames=usernames,
            priorities=priorities,
            criticities=criticities,
            message=message,
            classification=classification,
            description=description,
            greater_than_timestamp=greater_than_timestamp,
            less_than_timestamp=less_than_timestamp,
            timezone=timezone,
            page=page,
            limit=limit
            )

    def get_summary(self)->tuple[list, str]:
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
            
        return Events.serialize()
    
class EventsLoggerEngine(BaseEngine):
    r"""
    Data logger Engine class for Tag thread-safe database logging.

    """

    def __init__(self):

        super(EventsLoggerEngine, self).__init__()
        self.logger = EventsLogger()

    def create(
        self,
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        ):

        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["message"] = message
        _query["parameters"]["user"] = user
        _query["parameters"]["description"] = description
        _query["parameters"]["classification"] = classification
        _query["parameters"]["priority"] = priority
        _query["parameters"]["criticity"] = criticity
        _query["parameters"]["timestamp"] = timestamp
        
        return self.query(_query)
    
    def get_lasts(
        self,
        lasts:int=1
        ):

        _query = dict()
        _query["action"] = "get_lasts"
        _query["parameters"] = dict()
        _query["parameters"]["lasts"] = lasts
        
        return self.query(_query)
    
    def filter_by(
        self,
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        message:str="",
        classification:str="",
        description:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):

        _query = dict()
        _query["action"] = "filter_by"
        _query["parameters"] = dict()
        _query["parameters"]["usernames"] = usernames
        _query["parameters"]["priorities"] = priorities
        _query["parameters"]["criticities"] = criticities
        _query["parameters"]["message"] = message
        _query["parameters"]["classification"] = classification
        _query["parameters"]["description"] = description
        _query["parameters"]["greater_than_timestamp"] = greater_than_timestamp
        _query["parameters"]["less_than_timestamp"] = less_than_timestamp
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        
        return self.query(_query)

    def get_summary(self):
        r"""
        Documentation here
        """
        _query = dict()
        _query["action"] = "get_summary"
        _query["parameters"] = dict()
        
        return self.query(_query)
    


import pytz
from peewee import CharField, TimestampField, ForeignKeyField, IntegerField, fn
from ..dbmodels.core import BaseModel
from datetime import datetime
from .users import Users
from ..modules.users.users import User

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"


class Events(BaseModel):
    
    timestamp = TimestampField(utc=True)
    message = CharField(max_length=256)
    description = CharField(max_length=256, null=True)
    classification = CharField(max_length=128, null=True)
    priority = IntegerField(null=True)
    criticity = IntegerField(null=True)
    user = ForeignKeyField(Users, backref='events')

    @classmethod
    def create(
        cls, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        )->tuple:

        if not isinstance(user, User):

            return None, f"User {user} - {type(user)} must be an User Object"
        
        user = Users.read_by_username(username=user.username) 

        if not timestamp:

            timestamp = datetime.now(pytz.UTC)
        
        if not isinstance(timestamp, datetime):

            return None, f"Timestamp must be a datetime Object"

        query = cls(
            message=message,
            user=user,
            description=description,
            classification=classification,
            priority=priority,
            criticity=criticity,
            timestamp=timestamp
        )
        query.save()

        return query, f"Event creation successful"
    
    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Documentation here
        """
        events = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [event.serialize() for event in events]
    
    @classmethod
    def filter_by(
        cls, 
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        description:str="",
        message:str="",
        classification:str="",
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Documentation here
        """
        import math
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if usernames:
            subquery = Users.select(Users.id).where(Users.username.in_(usernames))
            query = query.join(Users).where(Users.id.in_(subquery))
        
        if priorities:
            query = query.where(cls.priority.in_(priorities))
            
        if criticities:
            query = query.where(cls.criticity.in_(criticities))
            
        if description:
            query = query.where(fn.LOWER(cls.description).contains(description.lower()))
            
        if message:
            query = query.where(fn.LOWER(cls.message).contains(message.lower()))
            
        if classification:
            query = query.where(fn.LOWER(cls.classification).contains(classification.lower()))
            
        if greater_than_timestamp:
            greater_than_timestamp = _timezone.localize(datetime.strptime(greater_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp > greater_than_timestamp)
            
        if less_than_timestamp:
            less_than_timestamp = _timezone.localize(datetime.strptime(less_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp < less_than_timestamp)
            
        query = query.order_by(cls.id.desc())

        total_records = query.count()
        
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1
        
        paginated_query = query.paginate(page, limit)
        
        data = [event.serialize() for event in paginated_query]
        
        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }
    
    @classmethod
    def get_comments(cls, id:int):
        r"""
        Documentation here
        """
        query = cls.read(id=id)

        return [comment.serialize() for comment in query.logs]

    def serialize(self)-> dict:
        from .. import TIMEZONE, MANUFACTURER, SEGMENT
        timestamp = self.timestamp
        if timestamp:

            timestamp = timestamp.astimezone(TIMEZONE)
            timestamp = timestamp.strftime(DATETIME_FORMAT)

        return {
            "id": self.id,
            "timestamp": timestamp,
            "user": self.user.serialize(),
            "message": self.message,
            "description": self.description,
            "classification": self.classification,
            "priority": self.priority,
            "criticity": self.criticity,
            "segment": SEGMENT,
            "manufacturer": MANUFACTURER,
            "has_comments": True if self.logs else False
        }
    




from math import ceil
import dash
from ...utils import find_differences_between_lists
import plotly.graph_objects as go
from ...tags import CVTEngine

cvt = CVTEngine()


def init_callback(app:dash.Dash):

    @app.callback(
        dash.Output('filter_tags_dropdown', 'options'),
        dash.Input('filter_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        if pathname=="/filter":
                
            tags_options = [tag["name"] for tag in app.automation.cvt.get_tags()]
            
            return tags_options
        
        return dash.no_update
    
    @app.callback(
        dash.Output('filter_trends_figure', 'figure'),
        dash.Input('timestamp-interval', 'n_intervals'),
        dash.State('filter_tags_dropdown', 'value'),
        prevent_initial_call=True
        )
    def fig_tags(n_intervals, values):
        r"""
        Documentation here
        """
        if values:

            fig = go.Figure()
            counter_axis = 0
            labels = dict()
            units = list()
            
            for tag_name in values:

                timestamp = app.automation.das.buffer[tag_name]["timestamp"]
                values = app.automation.das.buffer[tag_name]["values"]
                unit = app.automation.das.buffer[tag_name]["unit"]

                if unit not in units:
                    counter_axis += 1
                    units.append(unit)

                if counter_axis==1:

                    fig.add_trace(go.Scatter(x=timestamp, y=values, name=tag_name))
                    labels["yaxis"] =  {
                            "title": unit
                        }
                else:

                    fig.add_trace(go.Scatter(x=timestamp, y=values, name=tag_name, yaxis=f"y{counter_axis}"))
                    labels[f"yaxis{counter_axis}"] = {
                            "title": unit,
                            "anchor": "free",
                            "overlaying": "y",
                            "autoshift": True
                        }            

            fig.update_layout(**labels)

            return fig
        
        return dash.no_update
    

    @app.callback(
        dash.Output('filter_cvt_datatable', 'data', allow_duplicate=True),
        dash.Input('filter_page', 'pathname'),
        dash.Input('filter_tags_dropdown', 'value'),
        prevent_initial_call=True
        )
    def data_tags(pathname, values):
        r"""
        Documentation here
        """
        if values:

            data = list()
            for tag_name in values:

                tag = app.automation.cvt.get_tag_by_name(name=tag_name)

                data.append({
                    "id": tag.id,
                    "display_name": tag.display_name, 
                    "gaussian_filter": tag.gaussian_filter,
                    "threshold": tag.gaussian_filter_threshold,
                    "R-value": tag.gaussian_filter_r_value * 100.0
                })      

            return data
        
        return dash.no_update
    

    @app.callback(
        dash.Input('filter_last_values_dropdown', 'value'),
        dash.State('filter_tags_dropdown', 'value'),
        prevent_initial_call=True
        )
    def last_values(last_values, tags):
        r"""
        Documentation here
        """
        for tag_name in tags:

            # COMPUTATION OF MAX LENGTH OF THE BUFFER
            buffer_size = get_buffer_size(tag_name=tag_name, last_values=last_values)            
            app.automation.das.buffer[tag_name]["timestamp"].size = buffer_size
            app.automation.das.buffer[tag_name]["values"].size = buffer_size

    def get_buffer_size(tag_name:str, last_values:int):
        r"""
        Documentation here
        """
        tag = cvt.get_tag_by_name(name=tag_name)
        scan_time = tag.get_scan_time() # Milliseconds
        
        if not scan_time:
            
            current_timestamp = app.automation.das.buffer[tag_name]["timestamp"].current()
            previous_last = app.automation.das.buffer[tag_name]["timestamp"].previous_current()
            dt = current_timestamp - previous_last
            return ceil(last_values / dt.total_seconds())
        
        scan_time = scan_time / 1000
        return ceil(last_values / scan_time)
    

    @app.callback(
        dash.Output('filter_cvt_datatable', 'data'), 
        dash.Input('filter_cvt_datatable', 'data_timestamp'),
        dash.State('filter_cvt_datatable', 'data_previous'),
        dash.State('filter_cvt_datatable', 'data'),
        dash.State('filter_tags_dropdown', 'value'),
        )
    def update_tags(timestamp, previous, current, tags):
        
        message = None

        if timestamp:
            
            data = list()
            to_updates = find_differences_between_lists(previous, current)
            tag_to_update = to_updates[0]
            tag_id = tag_to_update.pop("id")
            
            if message:
                
                dash.set_props("modal-error-filter-tags-body", {"children": message})
                dash.set_props("modal-error-filter-tags", {'is_open': True})
                for tag_name in tags:

                    tag = app.automation.cvt.get_tag_by_name(name=tag_name)

                    data.append({
                        "id": tag.id,
                        "display_name": tag.display_name, 
                        "gaussian_filter": tag.gaussian_filter,
                        "threshold": tag.gaussian_filter_threshold,
                        "R-value": tag.gaussian_filter_r_value * 100.0
                    }) 
                
                return data

            if "segment" in tag_to_update:

                manufacturer_segment = tag_to_update['segment'].split("->")
                manufacturer = manufacturer_segment[0]
                segment = manufacturer_segment[1]
                tag_to_update.update({
                    "segment": segment,
                    "manufacturer": manufacturer
                })
            
            tag, message = app.automation.update_tag(id=tag_id, **tag_to_update)
            
            if not tag:

                dash.set_props("modal-error-filter-tags-body", {"children": message})
                dash.set_props("modal-error-filter-tags", {'is_open': True})
            
        for tag_name in tags:

            tag = app.automation.cvt.get_tag_by_name(name=tag_name)

            data.append({
                "id": tag.id,
                "display_name": tag.display_name, 
                "gaussian_filter": tag.gaussian_filter,
                "threshold": tag.gaussian_filter_threshold,
                "R-value": tag.gaussian_filter_r_value * 100.0
            }) 
        
        return data

    @app.callback(
        dash.Output("modal-update-filter-tags", "is_open"),
        dash.Input("close-modal-update-filter-tags", "n_clicks"),
        [dash.State("modal-update-filter-tags", "is_open")],
    )
    def close_error_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open
    
    @app.callback(
        dash.Output("modal-success-filter-tags", "is_open"),
        dash.Input("close-modal-success-filter-tags", "n_clicks"),
        [dash.State("modal-success-filter-tags", "is_open")],
    )
    def close_success_button(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open


import dash
import dash_bootstrap_components as dbc

class OPCUAServerComponents:

    @classmethod
    def opcua_server_table(cls)->dash.dash_table.DataTable:
        r"""
        Documentation here
        """
        return dbc.Container(   
            dbc.Row(
                dbc.Col(
                    dash.dash_table.DataTable(
                        data=[],
                        columns=[
                            {'name': 'name', 'id': 'name', 'editable': False}, 
                            {'name': 'namespace', 'id': 'namespace', 'editable': False}, 
                            {'name': 'access type', 'id': 'access_type', 'presentation': 'dropdown', 'clearable': False}
                        ],
                        id="opcua_server_datatable",
                        filter_action="native",
                        sort_action="native",
                        sort_mode="multi",
                        row_deletable=True,
                        selected_columns=[],
                        dropdown = {
                            'access_type': {
                                'options': [
                                    {'label': 'Read', 'value': 'Read'},
                                    {'label': 'ReadWrite', 'value': 'ReadWrite'},
                                    {'label': 'Write', 'value': 'Write'}
                                ],
                                'clearable': False
                            }
                        },
                        page_action="native",
                        page_current= 0,
                        page_size= 10,
                        persistence=True,
                        editable=True,
                        persisted_props=['data'],
                        export_format='xlsx',
                        export_headers='display',
                        style_table={'overflowX': 'auto'},
                    ),
                    width=12,
                )
            ),
            fluid=True,
            className="mx-0 px-0",
            
        )


import pywt
import numpy as np
from datetime import datetime

class Wavelet:
    
    def __init__(
        self, 
        pipeline_length:float, 
        speed_of_sound:float, 
        family:str="db2"
        ):

        # Constant definitions
        self.C = speed_of_sound                             # Sound speed constant, (m/s)
        self.L = pipeline_length                            # Pipeline length (m)
        self.FAMILY = family
        self.threshold_iqr = 95

    def set_speed_of_sound(self, value:float)->None:
        r"""
        Documentation here
        """
        self.C = value

    def get_speed_of_sound(self)->float:
        r"""
        Documentation here
        """
        return self.C

    def set_threshold_iqr(self, value:float):
        r"""
        Documentation here
        """
        if value < 0 or value > 100:
            raise ValueError("Threshold IQR must be between 0 and 100")
        self.threshold_iqr = value

    def get_threshold_iqr(self)->float:
        r"""
        Documentation here
        """
        return self.threshold_iqr

    def set_pipeline_length(self, value:float)->None:
        r"""
        Documentation here
        """
        self.L = value
    
    def get_pipeline_length(self)->float:
        r"""
        Documentation here
        """
        return self.L
    
    def set_family(self, value:str)->None:
        r"""
        Set Wavelet Family
        ['bior1.1', 'bior1.3', 'bior1.5', 'bior2.2', 'bior2.4', 'bior2.6', 'bior2.8', 'bior3.1', 'bior3.3', 
        'bior3.5', 'bior3.7', 'bior3.9', 'bior4.4', 'bior5.5', 'bior6.8', 'cgau1', 'cgau2', 'cgau3', 'cgau4', 
        'cgau5', 'cgau6', 'cgau7', 'cgau8', 'cmor', 'coif1', 'coif2', 'coif3', 'coif4', 'coif5', 'coif6', 
        'coif7', 'coif8', 'coif9', 'coif10', 'coif11', 'coif12', 'coif13', 'coif14', 'coif15', 'coif16', 
        'coif17', 'db1', 'db2', 'db3', 'db4', 'db5', 'db6', 'db7', 'db8', 'db9', 'db10', 'db11', 'db12', 
        'db13', 'db14', 'db15', 'db16', 'db17', 'db18', 'db19', 'db20', 'db21', 'db22', 'db23', 'db24', 
        'db25', 'db26', 'db27', 'db28', 'db29', 'db30', 'db31', 'db32', 'db33', 'db34', 'db35', 'db36', 
        'db37', 'db38', 'dmey', 'fbsp', 'gaus1', 'gaus2', 'gaus3', 'gaus4', 'gaus5', 'gaus6', 'gaus7', 
        'gaus8', 'haar', 'mexh', 'morl', 'rbio1.1', 'rbio1.3', 'rbio1.5', 'rbio2.2', 'rbio2.4', 'rbio2.6', 
        'rbio2.8', 'rbio3.1', 'rbio3.3', 'rbio3.5', 'rbio3.7', 'rbio3.9', 'rbio4.4', 'rbio5.5', 'rbio6.8', 
        'shan', 'sym2', 'sym3', 'sym4', 'sym5', 'sym6', 'sym7', 'sym8', 'sym9', 'sym10', 'sym11', 'sym12', 
        'sym13', 'sym14', 'sym15', 'sym16', 'sym17', 'sym18', 'sym19', 'sym20']
        """
        self.FAMILY = value

    def get_family(self)->str:
        r"""
        Get Wavelet Family
        """
        return self.FAMILY

    def detect_anomaly(self, data):
        r"""
        Documentation here
        """
        cA, cD = self.decomposition(data)
        lower_anomaly, probabilities = self.find_singular_points(cD)
        return lower_anomaly, probabilities
    
    def find_singular_points(self, cD):
        r"""
        Find singular points in the wavelet detail coefficients.
        """
        x = np.arange(len(cD))
        slope, _ = np.polyfit(x, cD, 1)

        # print(f"Slope: {slope}")
        threshold = np.percentile(cD, 100 - self.threshold_iqr)
        singular_points = np.where(cD < threshold)[0]
    
        # Calcular probabilidad basada en la magnitud de la caída
        if len(singular_points) > 0 and abs(slope) > 0.5:
            magnitudes = np.abs(cD[singular_points] - threshold)
            # Normalizar magnitudes a probabilidades entre 0 y 1
            probabilities = magnitudes / np.max(magnitudes)
            
            return singular_points, probabilities
        return [], []
    
    
    def map_to_original_position(self, anomaly, original_length, cD_length):
        r"""
        Map the indices of anomalies to the positions in the original data.
        """
        # Assuming level=1 for simplicity, adjust if using different levels
        original_position = None
        if anomaly:

            scaling_factor = anomaly / cD_length
            original_position = (scaling_factor * original_length).astype(int)

        return original_position

    def decomposition(self, data):
        r"""
        Documentation here
        """
        coeffs = pywt.wavedec(data, self.FAMILY, level=1)
        cA = coeffs[0]
        cD = coeffs[1]
        return cA, cD

    def wavedec(self, s, wavelet, mode='symmetric', level=None, axis=-1):
        r"""
        Multilevel 1D Discrete Wavelet Transform of signal $s$
        
        **Parameters**

        * **s:** (array_like) Input data
        * **wavelet:** (Wavelet object or name string) Wavelet to use
        * **mode:** (str) Signal extension mode.
        * **level:** (int) Decomposition level (must be >= 0). If level is None (default)
        then it will be calculated using the `dwt_max_level` function.
        * **axis:** (int) Axis over which to compute the DWT. If not given, the last axis 
        is used.

        **Returns**

        * **[cA_n, cD_n, cD_n-1, ..., cD2, cD1]:** (list) Ordered list of coefficients arrays where
        $n$ denotes the level of decomposition. The first element `(cA_n)` of the result is approximation
        coefficients array and the following elements `[cD_n - cD1]` are details coefficients arrays.

        ## Snippet code

        ```python
        >>> coeffs = wavedec([1,2,3,4,5,6,7,8], 'db1', level=2)
        >>> cA2, cD2, cD1 = coeffs
        >>> cD1
        array([-0.70710678, -0.70710678, -0.70710678, -0.70710678])
        >>> cD2
        array([-2., -2.])
        >>> cA2
        array([ 5., 13.])
        >>> s = np.array([[1,1], [2,2], [3,3], [4,4], [5, 5], [6, 6], [7, 7], [8, 8]])
        >>> coeffs = wavedec(s, 'db1', level=2, axis=0)
        >>> cA2, cD2, cD1 = coeffs
        >>> cD1
        array([[-0.70710678, -0.70710678],
               [-0.70710678, -0.70710678],
               [-0.70710678, -0.70710678],
               [-0.70710678, -0.70710678]])
        >>> cD2
        array([[-2., -2.],
               [-2., -2.]])
        >>> cA2
        array([[ 5.,  5.],
               [13., 13.]])

        ```
        """
        coeffs = pywt.wavedec(s, wavelet, mode=mode, level=level, axis=axis)

        return coeffs
    
    def time_flight(self, dt):
        """
        x = L/2 ± (C * dt)/2
        
        Donde:
        x: distancia desde el sensor de entrada hasta la fuga
        L: longitud total de la tubería
        C: velocidad del sonido en el fluido
        dt: tiempo entre detecciones (t2 - t1)
        El signo ± depende de qué sensor detectó primero la onda
        """
        if dt > 0:  # La onda llegó primero al sensor de entrada
            location = (self.L - self.C * abs(dt)) / 2
        else:  # La onda llegó primero al sensor de salida
            location = (self.L + self.C * abs(dt)) / 2

        return location

    def compute_mode(self, arr):
        unique_values, counts = np.unique(arr, return_counts=True)
        max_count = np.max(counts)
        modes = unique_values[counts == max_count]
        return modes


class NPW:
    r"""
    Documentation here
    """

    def __init__(self, pipeline_length: int, speed_of_sound: float=1180, family: str="db2"):
        
        self.wavelet = Wavelet(pipeline_length=pipeline_length, speed_of_sound=speed_of_sound, family=family)

    def detect_anomaly(self, *pressure_signals:list[np.ndarray]):
        r"""
        Documentation here
        """
        anomalies = list()
        probabilities = list()
        for signal in pressure_signals:
            anomaly, probability = self.wavelet.detect_anomaly(signal)
            
            if len(anomaly) >= 1:
                anomalies.append(True)
                probabilities.append(probability)
        return anomalies, probabilities
    
    def diagnose_anomaly(
            self, 
            inlet_pressure:list[float], 
            outlet_pressure:list[float],
            inlet_timestamps:list[datetime], 
            outlet_timestamps:list[datetime]
        ):
        r"""
        Documentation here
        """

        if not isinstance(inlet_timestamps, list) or not all(isinstance(x, datetime) for x in inlet_timestamps):
            
            raise ValueError("inlet_timestamps must be a list of datetime objects when diagnosing anomalies on NPW Algorithm")
        
        if not isinstance(outlet_timestamps, list) or not all(isinstance(x, datetime) for x in outlet_timestamps):
            
            raise ValueError("outlet_timestamps must be a list of datetime objects when diagnosing anomalies on NPW Algorithm")
        
        if not isinstance(inlet_pressure, list) or not all(isinstance(x, float) for x in inlet_pressure):
            
            raise ValueError("inlet_pressure must be a list of float objects when diagnosing anomalies on NPW Algorithm")
        
        if not isinstance(outlet_pressure, list) or not all(isinstance(x, float) for x in outlet_pressure):
            
            raise ValueError("outlet_pressure must be a list of float objects when diagnosing anomalies on NPW Algorithm")

        inlet_timestamp = None
        outlet_timestamp = None
        # Inlet Pressure
        _, cD = self.wavelet.decomposition(inlet_pressure)
        anomalies, _ = self.wavelet.find_singular_points(cD)
        if len(anomalies) >= 1:
            anomaly = anomalies[0]
            pos = self.wavelet.map_to_original_position(anomaly=anomaly, original_length=len(inlet_pressure), cD_length=len(cD))

            if pos:
                
                inlet_timestamp = inlet_timestamps[pos]

        # Outlet Pressure
        _, cD = self.wavelet.decomposition(outlet_pressure)
        anomalies, _ = self.wavelet.find_singular_points(cD)
        if len(anomalies) >= 1:
            anomaly = anomalies[0]
            pos = self.wavelet.map_to_original_position(anomaly=anomaly, original_length=len(outlet_pressure), cD_length=len(cD))
            if pos:
                outlet_timestamp = outlet_timestamps[pos]

        if inlet_timestamp and outlet_timestamp:
            
            dt = outlet_timestamp - inlet_timestamp
            location = self.wavelet.time_flight(dt=dt.total_seconds())
            if location < 0:
                
                return 0
            
            if location > self.wavelet.L:
                
                return self.wavelet.L
            
            return location



import secrets
from ...singleton import Singleton


class Role:
    r"""
    Documentation here
    """

    def __init__(self, name:str, level:int, identifier:str=None):
        
        _identifier = secrets.token_hex(4)
        
        if identifier:

            _identifier = identifier

        self.identifier = _identifier
        self.name:str = name
        self.level:int = level

    def serialize(self):
        r"""
        Documentation here
        """
        return {
            "identifier": self.identifier,
            "name": self.name,
            "level": self.level
        }
    

class Roles(Singleton):
    r"""
    Documentation here
    """

    def __init__(self):

        self.roles = dict()

    def add(self, role:Role)->str:
        r"""
        Documentation here
        """
        if isinstance(role, Role):

            if not self.check_role_name(name=role.name):

                self.roles[role.identifier] = role

                return role.identifier, f"role creation successful"
            
            return None, f"role {role.name} is already used"

        else:

            raise TypeError(f"{role} must be a Role instale")
        
    def get(self, id:str)->Role:
        r"""
        Documentation here
        """
        if id in self.roles:

            return self.roles[id]
        
    def get_by_name(self, name:str)->Role:
        r"""
        Documentation here
        """
        for _, role in self.roles.items():

            if name.lower()==role.name.lower():

                return role
            
    def get_names(self)->list:
        r"""
        Documentation here
        """
        return [role.name for _, role in self.roles.items()]
    
    def put(self, id:str, **kwargs):
        r"""
        Documentation here
        """
        if id in self.roles:
            role = self.roles[id]
            fields = {key: value for key, value in kwargs.items() if key in role.serialize()}
            self.roles[id].__dict__.update(fields)

    def delete(self, id:str)->dict:
        r"""
        Documentation here
        """
        if id in self.roles:
            
            return self.roles.pop(id)
        
    def _delete_all(self):

        self.roles = dict()
    
    def check_role_name(self, name:str):
        r"""
        Documentation here
        """
        if self.get_by_name(name=name):

            return True
        
        return False
    
    def serialize(self):
        r"""
        Documentation here
        """
        return [role.serialize() for _, role in self.roles.items() if role.name.lower()!="sudo"]
    
roles = Roles()


import os, pytz
from flask import Flask
from .core import PyAutomation
from .state_machine import OPCUAServer

app = Flask(__name__, instance_relative_config=False)

MANUFACTURER = os.environ.get('MANUFACTURER')
SEGMENT = os.environ.get('SEGMENT')
_TIMEZONE = os.environ.get('TIMEZONE') or "America/Caracas"
TIMEZONE = pytz.timezone(_TIMEZONE)
CERT_FILE = os.path.join(".", "ssl", os.environ.get('CERT_FILE') or "")
KEY_FILE = os.path.join(".", "ssl", os.environ.get('KEY_FILE') or "")
if not os.path.isfile(CERT_FILE):
    CERT_FILE = None

if not os.path.isfile(KEY_FILE):
    KEY_FILE = None
OPCUA_SERVER_PORT = os.environ.get('OPCUA_SERVER_PORT') or "53530"


class CreateApp():
    """Initialize the core application."""

    def __call__(self):
        """
        Documentation here
        """
        app.client = None
        self.application = app
        
        with app.app_context():

            from . import extensions
            extensions.init_app(app)

            from . import modules
            modules.init_app(app)
            
            return app
        
__application = CreateApp()
server = __application()    
server.config['TPT_TOKEN'] = '073821603fcc483f9afee3f1500782a4'
server.config['BUNDLE_ERRORS'] = True
opcua_server = OPCUAServer()



import dash

def init_callback(app:dash.Dash):

    @app.callback(
        dash.Output('alarms_summary_datatable', 'data'),
        dash.Input('alarms_history_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        if pathname=="/alarms-history":

            data = app.automation.get_lasts_alarms()

            return data
        
        return dash.no_update


import dash
import dash_bootstrap_components as dbc
from automation.pages.components.machines import MachinesComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dbc.Breadcrumb(
            items=[
                {"label": "Home", "href": "/"},  # Primer nivel
                {"label": "Machines", "active": True},  # Página actual (sin enlace)
            ],
        ),
        dash.dcc.Location(id='machines_page', refresh=False),
        MachinesComponents.machines_table()
    ],
    fluid=False,
    className="my-3",
)


import pytz
from peewee import CharField, TimestampField, ForeignKeyField, fn
from ..dbmodels.core import BaseModel
from datetime import datetime
from .users import Users
from .events import Events
from .alarms import AlarmSummary, Alarms
from ..modules.users.users import User

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"


class Logs(BaseModel):
    
    timestamp = TimestampField(utc=True)
    message = CharField(max_length=256)
    description = CharField(max_length=256, null=True)
    classification = CharField(max_length=128, null=True)
    user = ForeignKeyField(Users, backref='logs', on_delete='CASCADE')
    alarm = ForeignKeyField(AlarmSummary, null=True, backref='logs', on_delete='CASCADE')
    event = ForeignKeyField(Events, null=True, backref='logs', on_delete='CASCADE')

    @classmethod
    def create(
        cls, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        alarm_summary_id:int=None,
        event_id:int=None,
        timestamp:datetime=None
        )->tuple:
        if not isinstance(user, User):

            return None, f"User {user} - {type(user)} must be an User Object"
        
        _user = Users.read_by_username(username=user.username) 

        if not timestamp:

            timestamp = datetime.now()
        
        if not isinstance(timestamp, datetime):

            return None, f"Timestamp must be a datetime Object"
        
        query = cls(
            message=message,
            user=_user,
            description=description,
            classification=classification,
            timestamp=timestamp,
            event=Events.get_or_none(id=event_id),
            alarm=AlarmSummary.get_or_none(id=alarm_summary_id)
        )
        query.save()

        return query, f"Event creation successful"
    
    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Documentation here
        """
        logs = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [log.serialize() for log in logs]
    
    @classmethod
    def filter_by(
        cls, 
        usernames:list[str]=None,
        alarm_names:list[str]=None,
        event_ids:list[int]=None,
        description:str="",
        message:str="",
        classification:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC'
        ):
        r"""
        Documentation here
        """
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if usernames:
            subquery = Users.select(Users.id).where(Users.username.in_(usernames))
            query = query.join(Users).where(Users.id.in_(subquery))
        
        if event_ids:
            subquery = Events.select(Events.id).where(Events.id.in_(event_ids))
            query = query.join(Events).where(Events.id.in_(subquery))
        
        if alarm_names:
            subquery = Alarms.select(Alarms.id).where(Alarms.name.in_(alarm_names))
            alarm_subquery = AlarmSummary.select(AlarmSummary.id).join(Alarms).where(Alarms.id.in_(subquery))
            query = query.join(AlarmSummary).where(AlarmSummary.id.in_(alarm_subquery))
        
        if description:
            query = query.where(fn.LOWER(cls.description).contains(description.lower()))
        
        if message:
            query = query.where(fn.LOWER(cls.message).contains(message.lower()))
        
        if classification:
            query = query.where(fn.LOWER(cls.classification).contains(classification.lower()))
        
        if greater_than_timestamp:
            greater_than_timestamp = _timezone.localize(datetime.strptime(greater_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp > greater_than_timestamp)
        
        if less_than_timestamp:
            less_than_timestamp = _timezone.localize(datetime.strptime(less_than_timestamp, '%Y-%m-%d %H:%M:%S.%f')).astimezone(pytz.UTC)
            query = query.where(cls.timestamp < less_than_timestamp)
        
        query = query.order_by(cls.id.desc())

        if not query.exists():
            
            return []
        
        return [log.serialize() for log in query]

    def serialize(self)-> dict:
        from .. import MANUFACTURER, SEGMENT, TIMEZONE
        timestamp = self.timestamp
        if timestamp:
            timestamp = pytz.UTC.localize(timestamp).astimezone(TIMEZONE)
            timestamp = timestamp.strftime(DATETIME_FORMAT)

        _event = None
        if self.event:

            _event = self.event.serialize()

        _alarm = None
        if self.alarm:

            _alarm = self.alarm.serialize()

        return {
            "id": self.id,
            "timestamp": timestamp,
            "user": self.user.serialize(),
            "message": self.message,
            "description": self.description,
            "classification": self.classification,
            "event": _event,
            "alarm": _alarm,
            "segment": SEGMENT,
            "manufacturer": MANUFACTURER
        }


import dash
import dash_mantine_components as dmc
import dash_bootstrap_components as dbc
from automation.pages.components import Components
from automation.pages.components.opcua import OPCUAComponents

# Set React version to 18.2.0
dash._dash_renderer._set_react_version('18.2.0')
dash.register_page(__name__, path="/")
opcua_components = OPCUAComponents()


layout = dmc.MantineProvider(
    [
        dbc.Container(
            [
                dash.html.Div(
                    [
                        Components.modal_error(
                            title="Success",
                            modal_id="modal-success-opcua-connection",
                            button_close_id="close-success-opcua-connection",
                            body_id="modal-success-opcua-connection-body"
                        ),
                        Components.modal_error(
                            title="Error",
                            modal_id="modal-error-opcua-connection",
                            button_close_id="close-error-opcua-connection",
                            body_id="modal-error-opcua-connection-body"
                        ),
                        dbc.Row(
                            [
                                dbc.Col(
                                    [
                                        dash.html.H1("OPCUA")
                                    ],
                                    width=6,
                                    className="col-md-2"
                                ),

                                dbc.Col(
                                    [
                                        dash.html.Div(
                                            [
                                                dbc.Button(
                                                    "Create", 
                                                    color="info",
                                                    outline=True,
                                                    id="add_server_button",
                                                ),   
                                                dbc.Button(
                                                    "Remove",  
                                                    color="danger", 
                                                    outline=True,
                                                    disabled=False, 
                                                    id="remove_server_button"
                                                ),
                                                dbc.Button(
                                                    "Download",  
                                                    color="primary", 
                                                    outline=True,
                                                    disabled=False, 
                                                    id="download_node_info_button"
                                                ),   
                                            ],
                                            className="d-grid gap-2 d-flex justify-content-end",
                                        )
                                    ],
                                    # width={
                                    #     "size": 3, 
                                    #     "offset": 7
                                    # }
                                    width=6,
                                    className="col-md-10"
                                )
                            ]
                        )
                    ]
                ),

                dash.html.Div(
                    [
                        dbc.Row(
                            [
                                dbc.Col(
                                    [],
                                    id="server_tree",  
                                    width=12,
                                    className="col-md-2 mb-3"
                                ),
                                dbc.Col(
                                    opcua_components.data_access_view_table(), 
                                    id="data_access_view_table", 
                                    width=12,
                                    className="col-md-10"
                                )
                            ]
                        ),
                    ]
                ),

                OPCUAComponents.add_server(
                    title="Add Server", 
                    modal_id="add_server_modal", 
                    body_id="add_server_body_modal", 
                    ok_button_id="add_server_ok_button_modal", 
                    cancel_button_id="add_server_cancel_button_modal"
                ),
                OPCUAComponents.remove_server(
                    title="Remove Server", 
                    modal_id="remove_server_modal", 
                    body_id="remove_server_body_modal", 
                    ok_button_id="remove_server_ok_button_modal", 
                    cancel_button_id="remove_server_cancel_button_modal"
                ),
                OPCUAComponents.download_node_info(
                    title="Download Node Info", 
                    modal_id="download_node_info_modal", 
                    body_id="download_node_info_body_modal", 
                    ok_button_id="download_node_info_ok_button_modal", 
                    cancel_button_id="download_node_info_cancel_button_modal"
                ),
                dash.dcc.Download(id="download")
            ],
            fluid=False,
            className="my-3"
        ),
    ]
)


import dash
import dash_bootstrap_components as dbc
from automation.pages.components.trends import TrendsComponents

dash.register_page(__name__)

layout = dbc.Container(
    [
        dbc.Breadcrumb(
            items=[
                {"label": "Home", "href": "/"},
                {"label": "Tags", "href": "/tags"},
                {"label": "Trends", "active": True},
            ],
        ),
        dash.dcc.Location(id='trends_page', refresh=False),
        dbc.Row(
            [
                dbc.Col(TrendsComponents.tags(), width=6, className="col-sm-6 col-md-10"),
                dbc.Col(TrendsComponents.last_values(), width=6, className="col-sm-6 col-md-2 ps-0"),
            ],
            className="mb-3"
        ),
        dbc.Row(
            [
                dbc.Col(TrendsComponents.current_value_table(), width=12, className="col-sm-12 col-md-2 mb-2"),
                dbc.Col(TrendsComponents.plot(), width=12, className="col-sm-12 col-md-10"),
            ],
            className="mb-3"
        ),
    ],
    fluid=False,
    className="my-3"
)
    


# -*- coding: utf-8 -*-
"""pyhades/_singleton.py

This module implements a singleton pattern design, a base class for
inheritance in other classes
"""
from weakref import WeakValueDictionary

class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """
    _instances = WeakValueDictionary()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            # This variable declaration is required to force a
            # strong reference on the instance.
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]
    


class Singleton(metaclass=SingletonMeta):
    """
    The Singleton with simple class name.
    """


import functools, logging, sys
from ..modules.users.users import User, Users
from ..logger.events import EventsLoggerEngine


events_engine = EventsLoggerEngine()
users = Users()

def decorator(declared_decorator):
    """
    Create a decorator out of a function, which will be used as a wrapper
    """

    @functools.wraps(declared_decorator)
    def final_decorator(func=None, **kwargs):
        # This will be exposed to the rest of your application as a decorator
        def decorated(func):
            # This will be exposed to the rest of your application as a decorated
            # function, regardless how it was called
            @functools.wraps(func)
            def wrapper(*a, **kw):
                # This is used when actually executing the function that was decorated

                return declared_decorator(func, a, kw, **kwargs)
            
            return wrapper
        
        if func is None:
            
            return decorated
        
        else:
            # The decorator was called without arguments, so the function should be
            # decorated immediately
            return decorated(func)

    return final_decorator

def set_event(message:str, classification:str, priority:int, criticity:int, description:str="", force:bool=False):
    @decorator
    def wrapper(func, args, kwargs):
        from automation import PyAutomation
        app = PyAutomation()
        result = func(*args, **kwargs)
        
        if result:
        
            if "user" in kwargs:

                user = kwargs.pop('user')
                if isinstance(user, User):

                    _description = None

                    if isinstance(result, tuple):

                        _description = result[-1]
                    
                    event, _ = events_engine.create(
                        message=message,
                        description=_description,
                        classification=classification,
                        priority=priority,
                        criticity=criticity,
                        user=user
                    )
                    if app.sio:

                        app.sio.emit("on.event", data=event.serialize())
        else:
            if force:
                user = users.get_by_username(username="system")
                event, _ = events_engine.create(
                    message=message,
                    description=description,
                    classification=classification,
                    priority=priority,
                    criticity=criticity,
                    user=user
                )
                
                if app.sio:

                    app.sio.emit("on.event", data=event.serialize())

        return result

    return wrapper

@decorator
def put_alarm_state(func, args, kwargs):
    r"""
    Documentation here
    """
    from ..logger.alarms import AlarmsLoggerEngine
    from .. import TIMEZONE
    alarms_engine = AlarmsLoggerEngine()   
    result = func(*args, **kwargs)
    alarm = args[0]
    alarms_engine.put(
        id=alarm.identifier,
        state=alarm.state.state
    )
    if alarm.sio:
        if alarm.timestamp:
            timestamp = alarm.timestamp
            timestamp = timestamp.astimezone(TIMEZONE)
            alarm.timestamp = timestamp
            if alarm.ack_timestamp:
                ack_timestamp = alarm.ack_timestamp
                ack_timestamp = ack_timestamp.astimezone(TIMEZONE)
                alarm.ack_timestamp = ack_timestamp
        
        alarm.sio.emit("on.alarm", data=alarm.serialize())
        
    return result

def validate_types(**validations):
    
    if "output" in validations:

        _output = validations.pop('output')

        if _output is None:

            _output = type(None)

    def decorator(func):

        def wrapper(*args, **kwargs):
            
            for key, _data_type in kwargs.items():

                if key in validations:
                
                    if not isinstance(_data_type, validations[key]):
                        message = f"Expected Input {key} as {validations[key]}, but got {type(_data_type)} in {func}"
                        logging.error(message)
                        raise TypeError(message)
                    
                else:
                    message = f"You didn't define {key} argument to validate in {func}"
                    logger = logging.getLogger("pyautomation")
                    logger.error(message)
                    raise KeyError(message)

            # Call the wrapped function
            result = func(*args, **kwargs)

            # Validate the output type
            if _output:
                
                if isinstance(_output, tuple):
                    
                    for counter, expected in enumerate(_output):
                        
                        if not isinstance(result[counter], expected):

                            message = f"Expected output type ({counter}) {expected}, but got {type(result[counter])} in func {func}"
                            logger = logging.getLogger("pyautomation")
                            logger.error(message)
                            raise TypeError(message)
                        
                else:

                    if not isinstance(result, _output):
                        message = f"Expected output type {_output}, but got {type(result)} in func {func}"
                        logger = logging.getLogger("pyautomation")
                        logger.error(message)
                        raise TypeError(message)

            return result
        return wrapper
    return decorator

@decorator
def logging_error_handler(func, args, kwargs):
    r"""
    Documentation here
    """
    try:
                
        result = func(*args, **kwargs)
        return result

    except Exception as ex:

        trace = []
        tb = ex.__traceback__
        while tb is not None:
            trace.append({
                "filename": tb.tb_frame.f_code.co_filename,
                "name": tb.tb_frame.f_code.co_name,
                "lineno": tb.tb_lineno
            })
            tb = tb.tb_next
        msg = str({
            'type': type(ex).__name__,
            'message': str(ex),
            'trace': trace
        })
        logger = logging.getLogger("pyautomation")
        logger.error(msg=msg)

@decorator
def db_rollback(func, args, kwargs):
    try:
        self = args[0]
        result = func(*args, **kwargs)
        return result

    except Exception as e:
        _, _, e_traceback = sys.exc_info()
        e_message = str(e)
        e_line_number = e_traceback.tb_lineno
        logger = logging.getLogger("pyautomation")
        logger.warning(f"Rollback in [line {e_line_number}] {self.__class__.__name__}.{func.__name__} - {e_message}")
        conn = self._db.connection()
        conn.rollback()
        result = func(*args, **kwargs)

        return result


from peewee import CharField, BooleanField, FloatField, ForeignKeyField, IntegerField, TimestampField, BooleanField
from .core import BaseModel
from datetime import datetime


class Manufacturer(BaseModel):
    
    name = CharField(unique=True)

    @classmethod
    def create(cls, name:str)-> dict:
        r"""

        """
        if not cls.name_exist(name):

            query = cls(name=name)
            query.save()
            
            return query

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Verify is a name exist into database

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "name": self.name
        }
    

class Segment(BaseModel):
    
    name = CharField()
    manufacturer = ForeignKeyField(Manufacturer, backref='segments')

    @classmethod
    def create(cls, name:str, manufacturer:str)-> dict:
        r"""Documentation here
        """
        if Manufacturer.name_exist(name=manufacturer):
            
            manufacturer_obj = Manufacturer.read_by_name(name=manufacturer)
        
        else:
            
            manufacturer_obj = Manufacturer.create(name=manufacturer)

        segment_obj = Segment.select().where(Segment.name == name, Segment.manufacturer == manufacturer_obj).exists()
        
        if not segment_obj:
            query = cls(name=name, manufacturer=manufacturer_obj)
            query.save()  
            return query
        
        return Segment.read_by_name(name=name)

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""

        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "name": self.name,
            "manufacturer": self.manufacturer.serialize()
        }


class Variables(BaseModel):
    
    name = CharField(unique=True)

    @classmethod
    def create(cls, name:str)-> dict:
        r"""
        You can use Model.create() to create a new model instance. This method accepts keyword arguments, where the keys correspond 
        to the names of the model's fields. A new instance is returned and a row is added to the table.

        ```python
        >>> Variables.create(name='Pressure')
        {
            'message': (str)
            'data': (dict) {
                'name': 'pressure'
            }
        }
        ```
        
        This will INSERT a new row into the database. The primary key will automatically be retrieved and stored on the model instance.

        **Parameters**

        * **name:** (str), Industrial protocol name

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (dict) row serialized}

        """
        result = dict()
        data = dict()

        if not cls.name_exist(name):

            query = cls(name=name)
            query.save()
            
            message = f"{name} variable created successfully"
            data.update(query.serialize())

            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        message = f"{name} variable is already into database"
        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Verify is a name exist into database

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "name": self.name
        }


class Units(BaseModel):
    
    name = CharField(unique=True)
    unit = CharField(unique=True)
    variable_id = ForeignKeyField(Variables, backref='units', on_delete='CASCADE')

    @classmethod
    def create(cls, name:str, unit:str, variable:str)-> dict:
        r"""
        You can use Model.create() to create a new model instance. This method accepts keyword arguments, where the keys correspond 
        to the names of the model's fields. A new instance is returned and a row is added to the table.

        ```python
        >>> Variables.create(name='Pa', variable='Pressure')
        {
            'message': (str)
            'data': (dict) {
                'id': 1,
                'name': 'Pa',
                'variable': 'pressure'
            }
        }
        ```
        
        This will INSERT a new row into the database. The primary key will automatically be retrieved and stored on the model instance.

        **Parameters**

        * **name:** (str), Industrial protocol name

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (dict) row serialized}

        """
        result = dict()
        data = dict()
        name = name

        if not cls.name_exist(name):

            query_variable = Variables.read_by_name(variable)
            
            if query_variable is not None:

                variable_id = query_variable

                query = cls(name=name, unit=unit, variable_id=variable_id)
                query.save()
                
                message = f"{name} unit created successfully"
                data.update(query.serialize())

                result.update(
                    {
                        'message': message, 
                        'data': data
                    }
                )
                return result


            message = f"{variable} variable not exist into database"

            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        message = f"{name} unit is already into database"
        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query.serialize()
        
        return None

    @classmethod
    def read_by_unit(cls, unit:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(unit=unit)
        
        if query is not None:

            return query
        
        return None
    
    @classmethod
    def read_by_unit(cls, unit:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(unit=unit)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Verify is a name exist into database

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "name": self.name,
            "variable": self.variable_id.name,
            "unit": self.unit
        }


class DataTypes(BaseModel):
    
    name = CharField(unique=True)

    @classmethod
    def create(cls, name:str)-> dict:
        r"""
        You can use Model.create() to create a new model instance. This method accepts keyword arguments, where the keys correspond 
        to the names of the model's fields. A new instance is returned and a row is added to the table.

        ```python
        >>> Variables.create(name='Pressure')
        {
            'message': (str)
            'data': (dict) {
                'name': 'pressure'
            }
        }
        ```
        
        This will INSERT a new row into the database. The primary key will automatically be retrieved and stored on the model instance.

        **Parameters**

        * **name:** (str), Industrial protocol name

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (dict) row serialized}

        """
        result = dict()
        data = dict()

        if not cls.name_exist(name):

            query = cls(name=name)
            query.save()
            
            message = f"{name} DataType created successfully"
            data.update(query.serialize())

            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        message = f"{name} DataType is already into database"
        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def read_by_name(cls, name:str)->bool:
        r"""
        Get instance by its a name

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return query
        
        return None

    @classmethod
    def name_exist(cls, name:str)->bool:
        r"""
        Verify is a name exist into database

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(name=name)
        
        if query is not None:

            return True
        
        return False

    def serialize(self)-> dict:
        r"""
        Serialize database record to a jsonable object
        """

        return {
            "id": self.id,
            "name": self.name
        }


class Tags(BaseModel):
    
    identifier = CharField(unique=True)
    name = CharField(unique=True)
    unit = ForeignKeyField(Units, backref='tags')
    data_type = ForeignKeyField(DataTypes, backref='tags')
    segment = ForeignKeyField(Segment, backref='tags', null=True)
    description = CharField(null=True, max_length=256)
    display_name = CharField(unique=True)
    display_unit = ForeignKeyField(Units)
    opcua_address = CharField(null=True)
    node_namespace = CharField(null=True)
    scan_time = IntegerField(null=True)
    dead_band = FloatField(null=True)
    active = BooleanField(default=True)
    process_filter = BooleanField(default=False)
    gaussian_filter = BooleanField(default=False)
    gaussian_filter_threshold = FloatField(default=1.0)
    gaussian_filter_r_value = FloatField(default=0.0)
    out_of_range_detection = BooleanField(default=False)
    outlier_detection = BooleanField(default=False)
    frozen_data_detection = BooleanField(default=False)

    @classmethod
    def create(
        cls, 
        id:str,
        name:str, 
        unit:str,
        data_type:str,
        description:str,
        display_name:str,
        display_unit:str,
        opcua_address:str="",
        node_namespace:str="",
        segment:str="",
        manufacturer:str="",
        scan_time:int=0,
        dead_band:float=0.0,
        active:bool=True,
        process_filter:bool=False,
        gaussian_filter:bool=False,
        gaussian_filter_threshold:float=1.0,
        gaussian_filter_r_value:float=0.0,
        out_of_range_detection:bool=False,
        outlier_detection:bool=False,
        frozen_data_detection:bool=False
        ):
        r"""
        Documentation here
        """
        result = dict()
        message = f"{name} already exist into database"
        data = dict()
        _unit = Units.read_by_unit(unit=unit)
        _display_unit = Units.read_by_unit(unit=display_unit)
        _data_type = DataTypes.read_by_name(name=data_type)
        
        if not cls.name_exist(name):

            if not cls.display_name_exist(name):
                
                if _unit is not None and _display_unit is not None:

                    if _data_type is not None:

                        if segment and manufacturer:

                            segment_obj = Segment.create(name=segment, manufacturer=manufacturer)

                            if not segment_obj:

                                result.update(
                                    {
                                        'message': f"Duplicated {manufacturer}->{segment}", 
                                        'data': data
                                    }
                                )

                                return result

                            
                            query = cls(
                                identifier=id,
                                name=name, 
                                unit=_unit,
                                data_type=_data_type,
                                description=description,
                                display_name=display_name,
                                display_unit=_display_unit,
                                opcua_address=opcua_address,
                                node_namespace=node_namespace,
                                scan_time=scan_time,
                                dead_band=dead_band,
                                active=active,
                                process_filter=process_filter,
                                gaussian_filter=gaussian_filter,
                                gaussian_filter_threshold=gaussian_filter_threshold,
                                gaussian_filter_r_value=gaussian_filter_r_value,
                                out_of_range_detection=out_of_range_detection,
                                outlier_detection=outlier_detection,
                                frozen_data_detection=frozen_data_detection,
                                segment=segment_obj
                                )
                        else:
                            query = cls(
                                identifier=id,
                                name=name, 
                                unit=_unit,
                                data_type=_data_type,
                                description=description,
                                display_name=display_name,
                                display_unit=_display_unit,
                                opcua_address=opcua_address,
                                node_namespace=node_namespace,
                                scan_time=scan_time,
                                dead_band=dead_band,
                                active=active,
                                process_filter=process_filter,
                                gaussian_filter=gaussian_filter,
                                gaussian_filter_threshold=gaussian_filter_threshold,
                                gaussian_filter_r_value=gaussian_filter_r_value,
                                out_of_range_detection=out_of_range_detection,
                                outlier_detection=outlier_detection,
                                frozen_data_detection=frozen_data_detection
                                )
                        query.save()
                        message = f"{name} tag created successfully"
                        
                        data.update(query.serialize())

                        result.update(
                            {
                                'message': message, 
                                'data': data
                            }
                        )
                        
                        return result

                    message = f"{data_type} data type not exist into database"
                    result.update(
                        {
                            'message': message, 
                            'data': data
                        }
                    )
                    return result

                message = f"{unit} unit not exist into database"
                result.update(
                    {
                        'message': message, 
                        'data': data
                    }
                )
                return result
        
        else:

            if _unit is not None and _display_unit is not None:

                    if _data_type is not None:
                        tag, _ = cls.get_or_create(name=name)
                        payload = {
                            "unit":_unit,
                            "data_type":_data_type,
                            "description":description,
                            "display_name":display_name,
                            "display_unit":_display_unit,
                            "opcua_address":opcua_address,
                            "node_namespace":node_namespace,
                            "scan_time":scan_time,
                            "dead_band":dead_band,
                            "active": True
                        }
                        cls.put(id=tag.id, **payload)

                    message = f"{data_type} data type not exist into database"
                    result.update(
                        {
                            'message': message, 
                            'data': data
                        }
                    )
                    return result
            
            message = f"{unit} unit not exist into database"
            result.update(
                {
                    'message': message, 
                    'data': data
                }
            )
            return result

        result.update(
            {
                'message': message, 
                'data': data
            }
        )
        return result

    @classmethod
    def put(cls, id:int, **fields)-> dict:
        r""""
        Update a single record

        Once a model instance has a primary key, you UPDATE a field by its id. 
        The model's primary key will not change:
        """ 
             
        if cls.id_exists(id):
            
            if "unit" in fields:

                unit = fields["unit"]
                if isinstance(unit, str):
                    query = Units.read_by_unit(unit=unit)
                    if query:

                        fields["unit"] = query

            if "display_unit" in fields:

                display_unit = fields["display_unit"]
                if isinstance(display_unit, str):
                    query = Units.read_by_unit(unit=display_unit)
                    if query:

                        fields["display_unit"] = query

            if "data_type" in fields:

                data_type = fields["data_type"]
                if isinstance(data_type, str):
                    query = DataTypes.read_by_name(name=data_type)
                    if query:

                        fields["data_type"] = query

            if "segment" in fields:

                if "manufacturer" in fields:

                    segment = fields["segment"]
                    manufacturer = fields.pop("manufacturer")
                    if isinstance(segment, str) and isinstance(manufacturer, str):
                        manufacturer_obj = Manufacturer.get_or_none(name=manufacturer)
                        if manufacturer_obj:
                            query = Segment.get_or_none((Segment.name == segment) & (Segment.manufacturer == manufacturer_obj))

                            if query:
                                
                                fields["segment"] = query
            
            query = cls.update(**fields).where(cls.id == id)
            query.execute()
            return query

    @classmethod
    def read_by_name(cls, name):
        r"""
        Documentation here
        """
        return cls.get_or_none(name=name)

    @classmethod
    def read_by_names(cls, names):
        r"""
        Documentation here
        """
        query = cls.select().where(cls.name in names)
        return query

    @classmethod
    def name_exist(cls, name):
        r"""
        Documentation here
        """
        tag = cls.get_or_none(name=name)
        if tag is not None:

            return True
        
        return False
    
    @classmethod
    def display_name_exist(cls, name):
        r"""
        Documentation here
        """
        tag = cls.get_or_none(name=name)
        if tag is not None:

            return True
        
        return False

    def get_machines(self):

        return self.machines

    def serialize(self):
        r"""
        Documentation here
        """
        segment = ""
        manufacturer = ""
        if self.segment:

            segment = self.segment.serialize()
            manufacturer = segment["manufacturer"]["name"]
            segment = segment["name"]

        gaussian_filter_r_value = 0
        if hasattr(self, "gaussian_filter_r_value"):

            gaussian_filter_r_value = self.gaussian_filter_r_value

        gaussian_filter_threshold = 0
        if hasattr(self, "gaussian_filter_threshold"):

            gaussian_filter_threshold = self.gaussian_filter_threshold

        return {
            'id': self.identifier,
            'name': self.name,
            'unit': self.unit.unit,
            'data_type': self.data_type.name,
            'description': self.description,
            'display_name': self.display_name,
            'display_unit': self.display_unit.unit,
            'opcua_address': self.opcua_address,
            'node_namespace': self.node_namespace,
            'scan_time': self.scan_time,
            'dead_band': self.dead_band,
            'variable': self.unit.variable_id.name,
            'active': self.active,
            'process_filter': self.process_filter,
            'gaussian_filter': self.gaussian_filter,
            'gaussian_filter_threshold': gaussian_filter_threshold,
            'gaussian_filter_r_value': gaussian_filter_r_value,
            'out_of_range_detection': self.out_of_range_detection,
            'frozen_data_detection': self.frozen_data_detection,
            'outlier_detection': self.outlier_detection,
            'segment': segment,
            "manufacturer": manufacturer
        }


class TagValue(BaseModel):
    
    tag = ForeignKeyField(Tags, backref='values')
    unit = ForeignKeyField(Units, backref='values')
    value = FloatField()
    timestamp = TimestampField(utc=True)

    class Meta:
        indexes = (
            (('timestamp',), False),
        )

    @classmethod
    def create(
        cls, 
        tag:Tags,
        value:float,
        timestamp:datetime,
        unit=Units):
        r"""
        Documentation here
        """
        query = cls(
            tag=tag,
            value=value, 
            timestamp=timestamp,
            unit=unit
            )
        query.save()


# -*- coding: utf-8 -*-
"""pyhades/logger/users.py
"""
from ..modules.users.users import User
from ..dbmodels import Users, Roles
from .core import BaseEngine, BaseLogger
from ..utils.decorators import db_rollback


class UsersLogger(BaseLogger):

    def __init__(self):

        super(UsersLogger, self).__init__()

    @db_rollback
    def set_role(self, name:str, level:int, identifier:str):
        r"""
        Documentation here
        """
        return Roles.create(
            name=name,
            level=level,
            identifier=identifier
        )

    @db_rollback
    def set_user(
            self, 
            user:User
        ):
        r"""
        Documentation here
        """
        return Users.create(
            user=user
        )
    
    @db_rollback
    def login(
            self, 
            password:str,
            username:str="",
            email:str=""
        ):
        r"""
        Documentation here
        """
        return Users.login(
            password=password,
            username=username,
            email=email
        )
    
class UsersLoggerEngine(BaseEngine):
    r"""
    Users logger Engine class for Tag thread-safe database logging.

    """

    def __init__(self):

        super(UsersLoggerEngine, self).__init__()
        self.logger = UsersLogger()

    def set_role(self, name:str, level:int, identifier:str):

        _query = dict()
        _query["action"] = "set_role"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["level"] = level
        _query["parameters"]["identifier"] = identifier
        
        return self.query(_query)
    
    def set_user(self, user:User):

        _query = dict()
        _query["action"] = "set_user"
        _query["parameters"] = dict()
        _query["parameters"]["user"] = user
        
        return self.query(_query)
    
    def login(self,password:str, username:str="", email:str=""):

        _query = dict()
        _query["action"] = "login"
        _query["parameters"] = dict()
        _query["parameters"]["password"] = password
        _query["parameters"]["username"] = username
        _query["parameters"]["email"] = email
        
        return self.query(_query)
    


import dash
from ...tags.cvt import CVTEngine
from ...utils import find_differences_between_lists, generate_dropdown_conditional


tag_engine = CVTEngine()

def init_callback(app:dash.Dash):

    @app.callback(
        dash.Input("tag_alarm_input", "value"),
        dash.Input("alarm_name_input", "value"), 
        dash.Input("alarm_type_input", "value"),
        dash.Input("alarm_trigger_value_input", "value")
        )
    def create_tag(
        tag:str,
        name:str,  
        type:str, 
        trigger_value:int|float
        )->str:
        r"""
        Documentation here
        """
        if tag:
            _tag = tag_engine.get_tag_by_name(name=tag)
            dash.set_props("alarm_trigger_unit", {'children': _tag.get_display_unit()})

        
        if name and tag and type and trigger_value:
            
            dash.set_props("create_alarm_button", {'disabled': False})

        else:
            
            dash.set_props("create_alarm_button", {'disabled': True})

    @app.callback(
        dash.Output("modal-alarm-create", "is_open"),
        dash.Input("close-model-alarm-create", "n_clicks"),
        [dash.State("modal-alarm-create", "is_open")],
    )
    def toggle_modal(n, is_open):
        r"""
        Documentation here
        """
        if n:

            return not is_open
        
        return is_open

    @app.callback(
        dash.Output("alarm_description_input", "value"),
        dash.Input("alarm_description_radio_button", "value")
    )
    def enable_description(enable:bool):
        r"""
        Documentation here
        """
        dash.set_props("alarm_description_input", {'disabled': not enable})
        return ""

    @app.callback(
        dash.Output('alarms_datatable', 'data', allow_duplicate=True),
        dash.Output('tag_alarm_input', 'options'),
        dash.Output('alarms_datatable', 'dropdown'),
        dash.Output('alarm_type_input', 'options'),
        dash.Output('alarms_datatable', 'dropdown_conditional'),
        dash.Input('alarms_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        if pathname=="/alarms":

            data = app.alarms_table_data()

            dropdown_options_type = [
                {'label': 'HIGH-HIGH', 'value': 'HIGH-HIGH'},
                {'label': 'HIGH', 'value': 'HIGH'},
                {'label': 'LOW', 'value': 'LOW'},
                {'label': 'LOW-LOW', 'value': 'LOW-LOW'},
                {'label': 'BOOL', 'value': 'BOOL'}
            ]
            dropdown_options_tag = [{"label": tag["name"], "value": tag["name"]} for tag in app.automation.cvt.get_tags()]
            dropdown = {
                "alarm_type": {
                    "options": dropdown_options_type,
                    "clearable": False
                },
                "tag": {
                    "options": dropdown_options_tag,
                    "clearable": False
                }
            }

            return data, dropdown_options_tag, dropdown, dropdown_options_type, generate_dropdown_conditional()
        
        return dash.no_update, dash.no_update, dash.no_update, dash.no_update, dash.no_update
    
    @app.callback(
        dash.Output('alarms_datatable', 'data', allow_duplicate=True),
        dash.Input('create_alarm_button', 'n_clicks'),
        dash.State("tag_alarm_input", "value"),
        dash.State("alarm_name_input", "value"), 
        dash.State("alarm_description_input", "value"),
        dash.State("alarm_type_input", "value"),
        dash.State("alarm_trigger_value_input", "value"),
        prevent_initial_call=True
    )
    def CreateAlarmButton(
        btn1, 
        tag_name,
        alarm_name,
        alarm_description,
        alarm_type,
        trigger_value
        ):
        r"""
        Documentation here
        """
        if "create_alarm_button" == dash.ctx.triggered_id:
            
            alarm, message = app.automation.create_alarm(
                name=alarm_name,
                tag=tag_name,
                alarm_type=alarm_type,
                trigger_value=float(trigger_value),
                description=alarm_description
            )
            
            if not alarm:
                
                dash.set_props("modal-body-alarm-create", {"children": message})
                dash.set_props("modal-alarm-create", {'is_open': True})
            
            return app.alarms_table_data()
        
    @app.callback(
        dash.Input('alarms_datatable', 'data_timestamp'),
        dash.State('alarms_datatable', 'data_previous'),
        dash.State('alarms_datatable', 'data'),
        )
    def delete_update_alarms(timestamp, previous, current):

        if timestamp:

            if previous and current: # UPDATE TAG DEFINITION
                
                to_updates = find_differences_between_lists(previous, current)
                alarm_to_update = to_updates[0]
                alarm_id = alarm_to_update.pop("id")
                message = f"Do you want to update alarm {alarm_id} To {alarm_to_update}?"

                # OPEN MODAL TO CONFIRM CHANGES
                dash.set_props("modal-update-delete-alarm-body", {"children": message})
                dash.set_props("modal-update-delete-alarm", {'is_open': True})

    @app.callback(
        [
            dash.Output("modal-update-delete-alarm", "is_open"), 
            dash.Output('alarms_datatable', 'data'), 
            dash.Output('alarms_datatable', 'data_timestamp'),
            dash.Output("update-delete-alarm-yes", "n_clicks"),
            dash.Output("update-delete-alarm-no", "n_clicks")
        ],
        [dash.Input("update-delete-alarm-yes", "n_clicks"), dash.Input("update-delete-alarm-no", "n_clicks")],
        [
            dash.State('alarms_datatable', 'data_timestamp'),
            dash.State("modal-update-delete-alarm", "is_open"),
            dash.State('alarms_datatable', 'data_previous'),
            dash.State('alarms_datatable', 'data')
        ]
    )
    def toggle_modal_update_delete_alarm(yes_n, no_n, timestamp, is_open, previous, current):
        r"""
        Documentation here
        """
        
        if yes_n:
            
            if timestamp:
                
                if len(previous) > len(current): # DELETE ALARM

                    removed_rows = [row for row in previous if row not in current]
                    
                    for row in removed_rows:
                        _id = row['id']
                        message = app.automation.delete_alarm(id=_id)
                        
                        if message:
                            dash.set_props("modal-body-alarm-create", {"children": message})
                            dash.set_props("modal-alarm-create", {'is_open': True})
                        
                elif previous and current: # UPDATE TAG DEFINITION
                    to_updates = find_differences_between_lists(previous, current)
                    alarm_to_update = to_updates[0]
                    alarm_id = alarm_to_update.pop("id")
                    message = app.automation.update_alarm(id=alarm_id, **alarm_to_update)
                    
                    if message:
                        dash.set_props("modal-body-alarm-create", {"children": message})
                        dash.set_props("modal-alarm-create", {'is_open': True})
                
                return not is_open, app.alarms_table_data(), None, 0, 0
        
        elif no_n:
            
            return not is_open, app.alarms_table_data(), None, 0, 0

        else:

            return is_open, app.alarms_table_data(), None, 0, 0
        


# -*- coding: utf-8 -*-
"""pyhades/logger/alarms.py
"""
from datetime import datetime
from ..dbmodels import Alarms, AlarmSummary, AlarmTypes, AlarmStates
from .core import BaseEngine, BaseLogger
from ..alarms.trigger import TriggerType
from ..alarms.states import AlarmState
from ..utils.decorators import db_rollback


class AlarmsLogger(BaseLogger):

    def __init__(self):

        super(AlarmsLogger, self).__init__()

    @db_rollback
    def create_tables(self, tables):
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return
        
        self._db.create_tables(tables, safe=True)
        self.__init_default_alarms_schema()

    @db_rollback
    def __init_default_alarms_schema(self):
        r"""
        Documentation here
        """
        ## Alarm Types
        for alarm_type in TriggerType:

            AlarmTypes.create(name=alarm_type.value)

        ## Alarm States
        for alarm_state in AlarmState._states:
            name = alarm_state.state
            mnemonic = alarm_state.mnemonic
            condition = alarm_state.process_condition
            status = alarm_state.alarm_status
            AlarmStates.create(name=name, mnemonic=mnemonic, condition=condition, status=status)

    @db_rollback
    def create(
            self,
            id:str,
            name:str,
            tag:str,
            trigger_type:str,
            trigger_value:float,
            description:str):
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return 

        Alarms.create(
            identifier=id,
            name=name,
            tag=tag,
            trigger_type=trigger_type,
            trigger_value=trigger_value,
            description=description
        )

    @db_rollback
    def get_alarms(self):
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return list()
        
        alarms = Alarms.read_all()

        if alarms:

            return alarms
    
    @db_rollback
    def get_alarm_by_name(self, name:str)->Alarms|None:
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return None
        
        return Alarms.read_by_name(name=name)

    @db_rollback        
    def get_lasts(self, lasts:int=10):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return list()
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.read_lasts(lasts=lasts)
    
    @db_rollback
    def filter_alarm_summary_by(
            self,
            states:list[str]=None,
            names:list[str]=None,
            tags:list[str]=None,
            greater_than_timestamp:datetime=None,
            less_than_timestamp:datetime=None,
            timezone:str="UTC",
            page:int=1,
            limit:int=20
        ):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.filter_by(
            states=states,
            names=names,
            tags=tags,
            greater_than_timestamp=greater_than_timestamp,
            less_than_timestamp=less_than_timestamp,
            timezone=timezone,
            page=page,
            limit=limit
        )
    
    @db_rollback
    def put(
        self,
        id:str,
        name:str=None,
        tag:str=None,
        description:str=None,
        alarm_type:str=None,
        trigger_value:str=None,
        state:str=None
        ):
        if not self.check_connectivity():
            
            return None
        
        fields = dict()
        alarm = Alarms.read_by_identifier(identifier=id)
        if alarm:
            if name:
                fields["name"] = name
            if tag:
                fields["tag"] = tag
            if description:
                fields["description"] = description
            if alarm_type:
                alarm_type = AlarmTypes.read_by_name(name=alarm_type)
                fields["trigger_type"] = alarm_type
            if trigger_value:
                fields["trigger_value"] = trigger_value
            if state:
                alarm_state = AlarmStates.get_or_none(name=state)
                fields["state"] = alarm_state
            query = Alarms.put(
                id=alarm.id,
                **fields
            )

            return query

    @db_rollback
    def delete(self, id:str):
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return None
        
        alarm_state = AlarmStates.get_or_none(name="Out Of Service")
        alarm = Alarms.read_by_identifier(identifier=id)
        Alarms.put(
            id=alarm.id,
            state=alarm_state
        )

    @db_rollback
    def create_record_on_alarm_summary(self, name:str, state:str, timestamp:datetime, ack_timestamp:datetime=None):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if self.check_connectivity():
            
            AlarmSummary.create(name=name, state=state, timestamp=timestamp, ack_timestamp=ack_timestamp)

    @db_rollback
    def put_record_on_alarm_summary(self, name:str, state:str=None, ack_timestamp:datetime=None):
        r"""
        Documentation here
        """
        if not self.check_connectivity():
            
            return None
        
        if not self.is_history_logged:

            return None
        
        fields = dict()
        alarm = AlarmSummary.read_by_name(name=name)

        if alarm:

            if ack_timestamp:
                fields["ack_time"] = ack_timestamp
            if state:
                alarm_state = AlarmStates.get_or_none(name=state)
                fields["state"] = alarm_state

            if fields:
                query = AlarmSummary.put(
                    id=alarm.id,
                    **fields
                )

                return query

    @db_rollback
    def get_alarm_summary(self):
        r"""
        Documentation here
        """
        if not self.is_history_logged:

            return None
        
        if not self.check_connectivity():
            
            return list()
        
        return AlarmSummary.read_all()
    
    
class AlarmsLoggerEngine(BaseEngine):
    r"""
    Alarms logger Engine class for Tag thread-safe database logging.

    """

    def __init__(self):

        super(AlarmsLoggerEngine, self).__init__()
        self.logger = AlarmsLogger()

    def create(
        self,
        id:str,
        name:str,
        tag:str,
        trigger_type:str,
        trigger_value:float,
        description:str
        ):

        _query = dict()
        _query["action"] = "create"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["name"] = name
        _query["parameters"]["tag"] = tag
        _query["parameters"]["trigger_type"] = trigger_type
        _query["parameters"]["trigger_value"] = trigger_value
        _query["parameters"]["description"] = description
        
        return self.query(_query)
    
    def get_lasts(
        self,
        lasts:int=1
        ):

        _query = dict()
        _query["action"] = "get_lasts"
        _query["parameters"] = dict()
        _query["parameters"]["lasts"] = lasts
        
        return self.query(_query)
    
    def get_alarms(self):

        _query = dict()
        _query["action"] = "get_alarms"
        _query["parameters"] = dict()
        
        return self.query(_query)
    
    def get_alarm_by_name(self, name:str):

        _query = dict()
        _query["action"] = "get_alarm_by_name"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        
        return self.query(_query)
    
    def filter_alarm_summary_by(
        self,
        names:list[str]=None,
        states:list[int]=None,
        tags:list[int]=None,
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):

        _query = dict()
        _query["action"] = "filter_alarm_summary_by"
        _query["parameters"] = dict()
        _query["parameters"]["names"] = names
        _query["parameters"]["states"] = states
        _query["parameters"]["tags"] = tags
        _query["parameters"]["greater_than_timestamp"] = greater_than_timestamp
        _query["parameters"]["less_than_timestamp"] = less_than_timestamp
        _query["parameters"]["timezone"] = timezone
        _query["parameters"]["page"] = page
        _query["parameters"]["limit"] = limit
        
        return self.query(_query)
    
    def create_record_on_alarm_summary(self, name:str, state:str, timestamp:datetime, ack_timestamp:datetime=None):

        _query = dict()
        _query["action"] = "create_record_on_alarm_summary"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["state"] = state
        _query["parameters"]["timestamp"] = timestamp
        _query["parameters"]["ack_timestamp"] = ack_timestamp
        
        return self.query(_query)
    
    def put_record_on_alarm_summary(
        self,
        name:str,
        state:str=None,
        ack_timestamp:datetime=None
        ):
        _query = dict()
        _query["action"] = "put_record_on_alarm_summary"
        _query["parameters"] = dict()
        _query["parameters"]["name"] = name
        _query["parameters"]["state"] = state
        _query["parameters"]["ack_timestamp"] = ack_timestamp

        return self.query(_query)

    def put(
        self,
        id:str,
        name:str=None,
        tag:str=None,
        description:str=None,
        alarm_type:str=None,
        trigger_value:str=None,
        state:str=None
        ):
        _query = dict()
        _query["action"] = "put"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        _query["parameters"]["name"] = name
        _query["parameters"]["tag"] = tag
        _query["parameters"]["description"] = description
        _query["parameters"]["alarm_type"] = alarm_type
        _query["parameters"]["trigger_value"] = trigger_value
        _query["parameters"]["state"] = state

        return self.query(_query)

    def delete(self, id:str):
        r"""
        Documentation here
        """
        _query = dict()
        _query["action"] = "delete"
        _query["parameters"] = dict()
        _query["parameters"]["id"] = id
        return self.query(_query)

    def get_alarm_summary(self):
        r"""
        Documentation here
        """
        _query = dict()
        _query["action"] = "get_alarm_summary"
        _query["parameters"] = dict()
        
        return self.query(_query)

    def create_tables(self, tables):
        r"""
        Create default PyHades database tables

        ['TagTrend', 'TagValue']

        **Parameters**

        * **tables** (list) list of database model

        **Returns** `None`
        """
        self.logger.create_tables(tables)


from flask_cors import CORS
from ..singleton import Singleton


class Cors(Singleton):

    def __init__(self):
        
        self.app = None

    def init_app(self, app):
        r"""
        Documentation here
        """

        self.app = CORS(app, resources={r"/*": {"origins": "*"}})

        return app


from ..utils.units import *

class Mass(EngUnit):
    """Creates a mass object that can store a mass value and 
    convert between units of mass.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Mass Object]\n

    ```python
    >>> from automation.variables.mass import Mass
    >>> mass = Mass(value=1.0, unit="kg")
    >>> mass.value
    1.0
    >>> mass.unit
    'kg'
    >>> mass.convert(to_unit="lb")
    2.2046226218
    >>> Mass.convert_value(value=1.0, from_unit="kg", to_unit="g")
    1000.0
    >>> Mass.convert_values(values=[1.0, 10.0], from_unit="kg", to_unit="g")
    [1000.0, 10000.0]
    >>> mass.change_unit(unit="lb")
    2.2046226218
    >>> mass.unit
    'lb'
    >>> mass.get_value()
    [2.2046226218, 'lb']
    >>> print(mass)
    2.2046226218 lb
    >>> mass2 = Mass(value=3.0, unit="g")
    >>> mass_result = mass + mass2
    >>> print(mass_result)
    2.2112364896654 lb
    >>> mass_result = mass * mass2
    >>> print(mass_result)
    0.014581082713656917 lb
    >>> mass_result = mass / mass2
    >>> print(mass_result)
    333.33333333333337 lb
    >>> mass_result = mass ** mass2
    >>> print(mass_result)
    1.0052423283919558 lb

    ```
    """
    
    class Units(UnitSerializer):
        kg = 'kg'
        g = 'g'
        mg = 'mg'
        lb = 'lb'
        metricTon = 'metricTon'
        oz = 'oz'
        grain = 'grain'
        shortTon = 'shortTon'
        longTon = 'longTon'
        slug = 'slug'
    
    conversions = {
        'kg' : 1.0,
        'g' : 1000.0,
        'mg' : 1000000.0,
        'metricTon' : 1.0 / 1000.0,
        'lb' : 2.2046226218,
        'oz' : 35.274,
        'grain' : 2.2046226218 * 7000.0,
        'shortTon' : 1.0 / 907.185,
        'longTon' : 1.0 / 1016.047,
        'slug' : 1.0 / 14.5939029
    }

    def __init__(self, value, unit):

        if unit not in Mass.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Mass.Units.list()}")
        
        super(Mass, self).__init__(value=value, unit=unit)


from peewee import CharField, IntegerField
from ..dbmodels.core import BaseModel

class OPCUA(BaseModel):
    
    client_name = CharField(unique=True)
    host = CharField()
    port = IntegerField()

    @classmethod
    def create(cls, client_name:str, host:str, port:int):
        r"""
        You can use Model.create() to create a new model instance. This method accepts keyword arguments, where the keys correspond 
        to the names of the model's fields. A new instance is returned and a row is added to the table.

        ```python
        >>> AlarmsType.create(name='High-High')
        {
            'message': (str)
            'data': (dict) {
                'name': 'HIGH-HIGH'
            }
        }
        ```
        
        This will INSERT a new row into the database. The primary key will automatically be retrieved and stored on the model instance.

        **Parameters**

        * **name:** (str), Industrial protocol name

        **Returns**

        * **result:** (dict) --> {'message': (str), 'data': (dict) row serialized}

        """

        if not cls.client_name_exist(client_name):

            query = cls(client_name=client_name, host=host, port=port)
            query.save()

            return query
        
    @classmethod
    def get_by_client_name(cls, client_name:str):
        r"""
        Documentation here
        """
        return cls.get_or_none(client_name=client_name)
    
    @classmethod
    def client_name_exist(cls, client_name:str):
        r"""
        Verify is a name exist into database

        **Parameters**

        * **name:** (str) Variable name

        **Returns**

        * **bool:** If True, name exist into database 
        """
        query = cls.get_or_none(client_name=client_name)
        
        if query is not None:

            return True
        
        return False
    
    def serialize(self):
        r"""
        Documentation here
        """
        return {
            "client_name": self.client_name,
            "host": self.host,
            "port": self.port
        }



# -*- coding: utf-8 -*-
"""rackio/workers/logger.py

This module implements Logger Worker.
"""
import logging, time, datetime, os, shutil
from .worker import BaseWorker
from ..managers import DBManager
from ..opcua.models import Client
from ..logger.datalogger import DataLoggerEngine
from ..tags.cvt import CVTEngine
import sqlite3
from peewee import SqliteDatabase
from ..dbmodels.tags import TagValue
from ..dbmodels.alarms import AlarmSummary
from ..dbmodels.events import Events
from ..dbmodels.logs import Logs


class LoggerWorker(BaseWorker):

    def __init__(self, manager:DBManager, period:int=5.0):

        super(LoggerWorker, self).__init__()
        
        self._manager = manager
        self._period = period
        self.logger = DataLoggerEngine()
        self.cvt = CVTEngine()
        self.sqlite_db = None
        self.sqlite_db_name = None

    def sqlite_db_backup(self):
        if self.sqlite_db:
            file_size_mb = os.path.getsize(self.sqlite_db_name) / 1024 / 1024 
            if file_size_mb > 1 * 1024: # 1 Gb: 
                timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                name = self.sqlite_db_name.split(".db")[0]
                name = name.split(os.path.sep)[-1]
                backup_file = os.path.join(".", "db", "backups", f"{name}_{timestamp}.db")
                shutil.copy2(os.path.join(".", "db", "app.db"), backup_file)
                logger = logging.getLogger("pyautomation")
                logger.info(f"Backup creado: {backup_file}")
                # Empty TagValue 
                query = TagValue.delete()
                query.execute()
                # Empty Alarm Summary
                query = AlarmSummary.delete()
                query.execute()
                # Empty Events
                query = Events.delete()
                query.execute()
                # Empty Logs
                query = Logs.delete()
                query.execute()
                # Execute Vacuum to compact DB
                self.sqlite_db.close()
                conn = sqlite3.connect(self.sqlite_db_name)
                cur = conn.cursor()
                cur.execute("VACUUM;")
                conn.commit()
                conn.close()
                # Reopen DB connection
                from ..dbmodels import proxy
                self._db = self._manager.get_db()
                proxy.initialize(self._db)

        else:
            db = self.logger.logger.get_db()
            if db:
                if isinstance(db, SqliteDatabase):
                    self.sqlite_db = db
                    self.sqlite_db_name = db.database


    def check_opcua_connection(self):
        from automation import PyAutomation
        app = PyAutomation()
        if app.opcua_client_manager._clients:
            for _, opcua_client in app.opcua_client_manager._clients.items():

                if isinstance(opcua_client, Client):
                    
                    opcua_client.reconnect()
        else:
            
            app.load_opcua_clients_from_db()

    def get_tags_from_queue(self, _queue):
        from .. import SEGMENT, MANUFACTURER
        tags = list()
        while not _queue.empty():

            item = _queue.get(block=False)
            tag_name = item["tag"]
            tag = self.cvt.get_tag_by_name(name=tag_name)
            if tag:

                if tag.manufacturer==MANUFACTURER and tag.segment==SEGMENT:

                    value = item['value']
                    timestamp = item["timestamp"]
                    tags.append({"tag":tag_name, "value":value, "timestamp":timestamp})

                elif not MANUFACTURER and not SEGMENT:

                    value = item['value']
                    timestamp = item["timestamp"]
                    tags.append({"tag":tag_name, "value":value, "timestamp":timestamp})

        return tags

    def reconnect_to_db(self):
        from automation import PyAutomation
        app = PyAutomation()
        
        if self.db_reconnection:
            
            logging.critical("Trying reconnect to DB...")
        
        self.db_reconnection = False
        db_connected = app.reconnect_to_db()
                
        if db_connected:
            
            logging.critical("Reconnection successfully")
            self.db_reconnection = True

    def run(self):
        r"""
        Documentation here
        """       
        _queue = self._manager.get_queue()
        self.db_reconnection = True

        while True:

            if self.db_reconnection:

                db_connection = self.logger.logger.check_connectivity()
            
                if db_connection:
                    self.sqlite_db_backup()
                    tags = self.get_tags_from_queue(_queue=_queue)
            
                    if tags:
                        
                        self.logger.write_tags(tags=tags)
                    
                else:

                    self.reconnect_to_db()
            
            else:

                self.reconnect_to_db()

            self.check_opcua_connection()

            if self.stop_event.is_set():
                logging.critical("Alarm worker shutdown successfully!")
                break

            time.sleep(self._period)



from math import ceil
import dash
import plotly.graph_objects as go
from ...tags import CVTEngine

cvt = CVTEngine()


def init_callback(app:dash.Dash):

    @app.callback(
        dash.Output('trends_tags_dropdown', 'options'),
        dash.Input('trends_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        
        if pathname=="/trends":
                
            tags_options = [tag["name"] for tag in app.automation.cvt.get_tags()]
            
            return tags_options
        
        return dash.no_update
    
    @app.callback(
        dash.Output('trends_cvt_datatable', 'data'),
        dash.Output('trends_figure', 'figure'),
        dash.Input('timestamp-interval', 'n_intervals'),
        dash.State('trends_tags_dropdown', 'value'),
        prevent_initial_call=True
        )
    def tags(n_intervals, values):
        r"""
        Documentation here
        """
        if values:

            fig = go.Figure()
            counter_axis = 0
            labels = dict()
            data = list()
            units = list()
            
            for tag_name in values:

                timestamp = app.automation.das.buffer[tag_name]["timestamp"]
                values = app.automation.das.buffer[tag_name]["values"]
                unit = app.automation.das.buffer[tag_name]["unit"]

                current_value = values.current()
                if current_value:
                    data.append({
                        "tag": tag_name, "value": f"{current_value} {unit}"
                    })

                if unit not in units:
                    counter_axis += 1
                    units.append(unit)

                if counter_axis==1:

                    fig.add_trace(go.Scatter(x=timestamp, y=values, name=tag_name))
                    labels["yaxis"] =  {
                            "title": unit
                        }
                else:

                    fig.add_trace(go.Scatter(x=timestamp, y=values, name=tag_name, yaxis=f"y{counter_axis}"))
                    labels[f"yaxis{counter_axis}"] = {
                            "title": unit,
                            "anchor": "free",
                            "overlaying": "y",
                            "autoshift": True
                        }            

            fig.update_layout(**labels)

            return data, fig
        
        return dash.no_update, dash.no_update
    

    @app.callback(
        dash.Input('trends_last_values_dropdown', 'value'),
        dash.State('trends_tags_dropdown', 'value'),
        prevent_initial_call=True
        )
    def last_values(last_values, tags):
        r"""
        Documentation here
        """
        for tag_name in tags:

            # COMPUTATION OF MAX LENGTH OF THE BUFFER
            buffer_size = get_buffer_size(tag_name=tag_name, last_values=last_values)            
            app.automation.das.buffer[tag_name]["timestamp"].size = buffer_size
            app.automation.das.buffer[tag_name]["values"].size = buffer_size

    def get_buffer_size(tag_name:str, last_values:int):
        r"""
        Documentation here
        """
        tag = cvt.get_tag_by_name(name=tag_name)
        scan_time = tag.get_scan_time() # Milliseconds
        
        if not scan_time:
            
            current_timestamp = app.automation.das.buffer[tag_name]["timestamp"].current()
            previous_last = app.automation.das.buffer[tag_name]["timestamp"].previous_current()
            dt = current_timestamp - previous_last
            return ceil(last_values / dt.total_seconds())
        
        scan_time = scan_time / 1000
        return ceil(last_values / scan_time)
        


# -*- coding: utf-8 -*-
"""automation/models.py

This module implements a Tags and other classes for
modelling the subjects involved in the core of the engine.
"""
from .modules.users.users import User
from .tags.tag import Tag
from .utils.decorators import set_event, logging_error_handler
from datetime import datetime, timezone


FLOAT = "float"
INTEGER = "int"
BOOL = "bool"
STRING = "str"

class PropertyType:

    """
    Implement an abstract property type
    """

    def __init__(self, _type, default=None, unit=None):
        from .tags.cvt import CVTEngine
        from .opcua.subscription import DAS
        self._type = _type
        self.unit = unit
        self.__value = default
        self.cvt = CVTEngine()
        self.das = DAS()

    @property
    def value(self):
        r"""
        Documentation here
        """
        return self.__value

    @value.setter
    def value(self, value):
        
        self.__value = value
    
    @set_event(message=f"Attribute updated", classification="State Machine", priority=2, criticity=3)
    def set_value(self, value, user:User=None, name:str=None, machine=None):
        
        if isinstance(self, ProcessType):

            if not self.read_only:

                if hasattr(self, "tag"):

                    if self.tag:

                        if hasattr(machine, "data_timestamp"):
                            timestamp = machine.data_timestamp
                        else:
                            timestamp = datetime.now(timezone.utc)
                        val = self.tag.value.convert_value(value=value.value, from_unit=self.tag.get_unit(), to_unit=self.tag.get_display_unit())
                        self.tag.value.set_value(value=val, unit=self.tag.get_display_unit()) 
                        self.cvt.set_value(id=self.tag.id, value=val, timestamp=timestamp)
                        if self.tag.get_name() in self.das.buffer:
                            self.das.buffer[self.tag.get_name()]["timestamp"](timestamp)
                            self.das.buffer[self.tag.get_name()]["values"](val)
        
        if self.value:
            if value.value!=self.value.value:
                if machine:
                    if machine.sio:
                        if name:
                            payload = {machine.name.value: {name: value.value}}
                            machine.sio.emit("on.machine.property", data=payload)
                            machine.sio.emit("on.machine", data=machine.serialize())

        if name=="machine_interval":
            
            return value, f"{name} To: {value.value} s."
        
        self.value = value
        
        return value, f"{name} To: {value.value}"


class StringType(PropertyType):

    """
    Implement a Float Type
    """

    def __init__(self, default=None, unit=None):

        super(StringType, self).__init__(STRING, default, unit)


class FloatType(PropertyType):

    """
    Implement a Float Type
    """

    def __init__(self, default=None, unit=None):

        super(FloatType, self).__init__(FLOAT, default, unit)


class IntegerType(PropertyType):

    """
    Implement an Integer Typle
    """

    def __init__(self, default=None, unit=None):

        super(IntegerType, self).__init__(INTEGER, default, unit)

        
class BooleanType(PropertyType):

    """
    Implement a Boolean Type
    """

    def __init__(self, default=None, unit=None):

        super(BooleanType, self).__init__(BOOL, default, unit)


class ProcessType(FloatType):

    """
    Implement a Process Type

    Attributes

    - read_only: [bool] only read from CVT, used to field data.
    - tag: [Tag] Tag binded on CVT
    """

    def __init__(self, tag:Tag|None=None, default=None, read_only:bool=True, unit:str=None):
        
        self.tag = tag
        self.read_only = read_only
        
        super(ProcessType, self).__init__(default=default, unit=unit)
        
    @logging_error_handler
    def serialize(self):
        r"""
        Documentation here
        """
        tag = None
        if self.tag:

            tag = self.tag.serialize()

        value = None
        if self.value:
            
            if isinstance(self.value, (bool, float, int, str)):

                value = self.value

            elif isinstance(self.value, (BooleanType, FloatType, IntegerType, StringType)):

                value = self.value.value

        return {
            "value": value,
            "unit": self.unit,
            "tag": tag,
            "read_only": self.read_only
        }





def assert_dict_contains_subset(subset, superset, msg=None):
        """
        Custom assertion to check if `subset` is a subset of `superset`.
        """
        missing_keys = {key for key in subset if key not in superset}
        assert not missing_keys, f"{msg or 'Dictionary subset check failed'}: Missing keys {missing_keys}"

        for key, value in subset.items():
            assert superset[key] == value, f"{msg or 'Dictionary subset check failed'}: Value mismatch for key '{key}'"


from ..utils.units import *

class Volume(EngUnit):
    """Creates a flow object that can store a flow value and 
    convert between units of flow.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Volume Object]\n

    ```python
    >>> from automation.variables.volume import Volume
    >>> volume = Volume(value=1.0, unit="bbl")
    >>> volume.value
    1.0
    >>> volume.unit
    'bbl'
    >>> volume.convert(to_unit="m3")
    0.158987294928
    >>> Volume.convert_value(value=1.0, from_unit="gal", to_unit="bbl")
    0.023809523809523808
    >>> Volume.convert_values(values=[1.0, 10.0], from_unit="gal", to_unit="bbl")
    [0.023809523809523808, 0.23809523809523808]
    >>> volume.change_unit(unit="gal")
    42.0
    >>> volume.unit
    'gal'
    >>> volume.get_value()
    [42.0, 'gal']
    >>> print(volume)
    42.0 gal
    >>> volume2 = Volume(value=120.0, unit="lt")
    >>> volume_result = volume + volume2
    >>> print(volume_result)
    73.7006462829778 gal
    >>> volume_result = volume * volume2
    >>> print(volume_result)
    1331.4271438850678 gal
    >>> volume_result = volume / volume2
    >>> print(volume_result)
    1.3248941244 gal
    >>> volume_result = volume ** volume2
    >>> print(volume_result)
    2.8711215836780378e+51 gal

    ```
    """

    class Units(UnitSerializer):
        bbl = 'bbl'
        gal = 'gal'
        cubic_meter = 'm3'
        liter_sec = 'lt'
        milliliter = 'ml'

    conversions = {
        'bbl' : 1.0,
        'gal' : 42.0,
        'm3' : 0.158987294928,
        'lt' : 158.987294928,
        'ml' : 158987.294928
    }

    def __init__(self, value, unit):

        if unit not in Volume.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Volume.Units.list()}")
        
        super(Volume, self).__init__(value=value, unit=unit)


from ..utils.units import *

class Time(EngUnit):
    """Creates a eng_time object that can store a eng_time value and 
    convert between units of eng_time.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Time Object]\n

    ```python
    >>> from automation.variables.eng_time import Time
    >>> eng_time = Time(value=1.0, unit="hr")
    >>> eng_time.value
    1.0
    >>> eng_time.unit
    'hr'
    >>> eng_time.convert(to_unit="minute")
    60.0
    >>> Time.convert_value(value=1.0, from_unit="hr", to_unit="s")
    3600.0
    >>> Time.convert_values(values=[1.0, 10.0], from_unit="hr", to_unit="s")
    [3600.0, 36000.0]
    >>> eng_time.change_unit(unit="minute")
    60.0
    >>> eng_time.unit
    'minute'
    >>> eng_time.get_value()
    [60.0, 'minute']
    >>> print(eng_time)
    60.0 minute
    >>> eng_time2 = Time(value=3.0, unit="s")
    >>> eng_time_result = eng_time + eng_time2
    >>> print(eng_time_result)
    60.05 minute
    >>> eng_time_result = eng_time * eng_time2
    >>> print(eng_time_result)
    3.0 minute
    >>> eng_time_result = eng_time / eng_time2
    >>> print(eng_time_result)
    1200.0 minute

    ```

    """
    
    class Units(UnitSerializer):
        ms = 'ms'
        s = 's'
        minute = 'minute'
        hr = 'hr'
        day = 'day'
    
    conversions = {
        'ms' : 1000.0,
        's' : 1.0,
        'minute' : 1.0 / 60.0,
        'hr' : 1.0 / 60.0 / 60.0,
        'day' : 1.0 / 60.0 / 60.0 / 24.0
    }

    def __init__(self, value, unit):

        if unit not in Time.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Time.Units.list()}")
        
        super(Time, self).__init__(value=value, unit=unit)


import dash
import dash_bootstrap_components as dbc

class DatabaseComponents:

    @classmethod
    def create_db_config_form(cls):
        r"""
        Documentation here
        """
        return dash.html.Div(
            [
                dash.dcc.Location(id='database_page', refresh=False),
                dbc.Accordion(
                [

                        dbc.AccordionItem(
                            [
                                dbc.Row(
                                    [
                                        dbc.Col([
                                            dbc.InputGroup(
                                                [
                                                    dbc.InputGroupText("Type"),
                                                    dbc.Select(
                                                        options=[
                                                            {"label": "postgresql", "value": "postgresql"},
                                                            {"label": "mysql", "value": "mysql"},
                                                            {"label": "sqlite", "value": "sqlite"}
                                                        ],
                                                        id="db_type_input"
                                                    ),
                                                    
                                                ],
                                                size="md"
                                            )
                                        ],
                                        width=12,
                                        className="col-sm-12 col-md-2"
                                        ),
                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="DB Name", id="db_name_input")
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),

                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="DB Host", id="db_host_input")
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),

                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="DB Port", id="db_port_input", type="number")
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-1"
                                        ),

                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(placeholder="DB User", id="db_user_input")
                                                    ], 
                                                    size="md"
                                                )
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-2"
                                        ),

                                        dbc.Col(
                                            [
                                                dbc.InputGroup(
                                                    [
                                                        dbc.Input(type="password", placeholder="DB Password", id="db_password_input")
                                                    ], 
                                                    size="md")
                                            ],
                                            width=12,
                                            className="col-sm-12 col-md-3"
                                        ),

                                        dbc.Col(
                                            dbc.Button(
                                                "Connect",
                                                color="primary",
                                                outline=True,
                                                disabled=True,
                                                id="connect_disconnect_db_button",
                                                className="w-100"
                                            ),
                                            width="auto",
                                            className="d-flex justify-content-center align-items-center"
                                        ),
                                    ],
                                    className="form g-3" 
                                ),
                            ],
                            title="Connection Settings",
                            className="my-3"
                        )
                    ],
                start_collapsed=False,
                )
            ]
        )



from enum import Enum

class UnitError(Exception):
    pass

class UnitSerializer(Enum):

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))
    
    @classmethod
    def serialize(cls):

        return {unit.name: unit.value for unit in cls}

class EngUnit(object):
    """Generic class for engineering unit objects containing a float value and string unit."""
    
    numerator = []
    denominator = []
    conversions = dict()
    
    def __init__(self, value, unit):
        super().__init__()
        self.value = value
        self.unit = unit
        self.baseUnit = dict(zip(self.conversions.values(), self.conversions.keys()))[1]

    def convert(self, to_unit):
        """Converts the object from one unit to another."""
        from_unit = self.unit
        to_unit = to_unit
        return float(self.value) / float(self.conversions[from_unit]) * float(self.conversions[to_unit])
    
    @classmethod
    def convert_values(self, values:list, from_unit:str, to_unit:str)->list:
        r"""
        Documentation here
        """
        return [float(value) / float(self.conversions[from_unit]) * float(self.conversions[to_unit]) for value in values]
    
    @classmethod
    def convert_value(cls, value:int|float, from_unit:str, to_unit:str)->float:
        """Unit value conversion

        :param value: [int|float] Value to convert
        :param from_unit: [str] Value's unit
        :param to_unit: [str] Unit which you want to convert the value
        :return: [float] Converted value into "to_unit"

        ```python
        >>> from automation.variables.pressure import Pressure
        >>> Pressure.convert_value(value=2, from_unit="atm", to_unit="Pa")
        202650.05476617732
        
        ```
        """
        return float(value) / float(cls.conversions[from_unit]) * float(cls.conversions[to_unit])
       
    def change_unit(self, unit):
        """Converts the current value of the object to a new unit.  Returns a float of the new value."""
        self.value = self.convert(unit)
        self.unit = unit
        return float(self.value)

    def set_value(self, value, unit):
        """Sets the value and unit of the object"""
        self.value = value
        self.unit = unit

    def get_value(self):
        """Returns a list of the float value and unit of the object."""
        return [float(self.value), self.unit]

    def __str__(self):
        return str(self.value) + ' ' + self.unit

    def __add__(self, other):
        new_value = self.value + other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __sub__(self, other):
        new_value = self.value - other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __mul__(self, other):
        new_value = self.value * other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __rmul__(self, other):
        new_value = self.value * other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __truediv__(self, other):
        new_value = self.value / other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __floordiv__(self, other):
        new_value = self.value // other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)

    def __pow__(self, other):
        new_value = self.value ** other.change_unit(self.unit)
        return self.__class__(new_value, self.unit)
    
    def __lt__(self, other):
        return self.value < other.change_unit(self.unit)

    def __le__(self, other):
        return self.value <= other.change_unit(self.unit)

    def __gt__(self, other):
        return self.value > other.change_unit(self.unit)

    def __ge__(self, other):
        return self.value >= other.change_unit(self.unit)
    


import unittest
from ..variables import (Pressure)

class TestConversions(unittest.TestCase):

    def setUp(self) -> None:
        
        return super().setUp()

    def tearDown(self) -> None:
        
        return super().tearDown()
    
    def test_pressure_conversions(self):

        value = 10
        from_unit = "atm"
        to_unit = "psi"
        expected = 146.959

        self.assertAlmostEqual(Pressure.convert_value(value, from_unit=from_unit, to_unit=to_unit), expected, delta=0.001)



import dash

def init_callback(app:dash.Dash):

    @app.callback(
        dash.Output('machines_datatable', 'data'),
        dash.Input('machines_page', 'pathname'),
        prevent_initial_call=True
        )
    def display_page(pathname):
        r"""
        Documentation here
        """
        if pathname=="/machines":

            data = app.machines_table_data()
            return data
        
        return dash.no_update
    
    @app.callback(
        dash.Output('machines_datatable', 'data', allow_duplicate=True),
        dash.Input('timestamp-interval', 'n_intervals'),
        prevent_initial_call=False
    )
    def update_table( n_intervals):
        
        data = app.machines_table_data()
        return data


from ..utils.units import EngUnit, UnitSerializer, UnitError

class Length(EngUnit):
    """Creates a length object that can store a length value and 
    convert between units of length.

    :param value: [int|float] Engineering value\n
    :param unit: [str] Engineering unit\n
    :return: [Length Object]\n

    ```python
    >>> from automation.variables.length import Length
    >>> length = Length(value=1.0, unit="m")
    >>> length.value
    1.0
    >>> length.unit
    'm'
    >>> length.convert(to_unit="inch")
    39.3701
    >>> Length.convert_value(value=1.0, from_unit="m", to_unit="cm")
    100.0
    >>> Length.convert_values(values=[1.0, 10.0], from_unit="m", to_unit="cm")
    [100.0, 1000.0]
    >>> length.change_unit(unit="inch")
    39.3701
    >>> length.unit
    'inch'
    >>> length.get_value()
    [39.3701, 'inch']
    >>> print(length)
    39.3701 inch
    >>> length2 = Length(value=3.0, unit="cm")
    >>> length_result = length + length2
    >>> print(length_result)
    40.551203 inch
    >>> length_result = length * length2
    >>> print(length_result)
    46.500143220300004 inch
    >>> length_result = length / length2
    >>> print(length_result)
    33.333333333333336 inch
    >>> length_result = length ** length2
    >>> print(length_result)
    76.56952620577046 inch
    
    ```
    """
    
    class Units(UnitSerializer):
        fm = 'fm' 
        pm = 'pm' 
        nm = 'nm' 
        um = 'um' 
        mm = 'mm' 
        cm = 'cm'
        m = 'm'
        dam = 'dam'
        hm = 'hm' 
        km = 'km'
        Mm = 'Mm' 
        Gm = 'Gm'
        Tm = 'Tm'
        Pm = 'Pm'

        inch = 'inch'
        ft = 'ft'
        yd = 'yd'
        mi = 'mi'

        nautMi = 'nautMi'
        lightYear = 'lightYear'

    conversions = {
        'fm' : 1000000000000000,
        'pm' : 1000000000000,
        'nm' : 1000000000,
        'um' : 1000000,
        'mm' : 1000,
        'cm' : 100,
        'm' : 1.0,
        'dam' : 0.1,
        'hm' : 0.01,
        'km' : 0.001,
        'Mm' : 0.000001,
        'Gm' : 0.000000001,
        'Tm' : 0.000000000001,
        'Pm' : 0.000000000000001,

        'inch' : 39.3701,
        'ft' : 3.28084,
        'yd' : 1.09361,
        'mi' : 0.000621371,

        'nautMi' : 1.0 / 1852.0,
        'lightYear' : 1.0 / (9.4607304725808 * (10**15))
    }

    def __init__(self, value, unit):

        if unit not in Length.Units.list():

            raise UnitError(f"{unit} value is not allowed for {self.__class__.__name__} object - you can use: {Length.Units.list()}")
        
        super(Length, self).__init__(value=value, unit=unit)


from flask_restx import Namespace, Resource
from .... import PyAutomation
from ....modules.users.roles import roles
from ....extensions.api import api
from ....extensions import _api as Api
from .models.roles import create_role_parser


ns = Namespace('Roles', description='Roles')
app = PyAutomation()

@ns.route('/')
class UsersByRoleResource(Resource):

    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    def get(self):
        """View all tecnogiros's role"""

        return roles.serialize(), 200

@ns.route('/add')
class CreateRoleResource(Resource):
    
    @Api.validate_reqparser(reqparser=create_role_parser)
    @api.doc(security='apikey')
    @Api.token_required(auth=True)
    @ns.expect(create_role_parser)
    def post(self):
        """Add Role"""  
        args = create_role_parser.parse_args()
        role, message = app.set_role(**args)
        
        if role:

            return role.serialize(), 200
        
        return message, 400